import re
import sys


def gtf_to_gene_info(gtf_file):
    fout = open(gtf_file + '.gene_info', 'w')
    fin = open(gtf_file)
    for line in fin:
        if '#' in line:
            continue
        arr = line.split()
        if arr[2] == 'gene':
            searchObj = re.search('gene_id\s+"(.+?)".+gene_name\s+"(.+?)"', line, re.I)
            gene_id = searchObj.group(1)
            gene_name = searchObj.group(2)
            gene_info = ' '.join([arr[0], arr[3], arr[4], arr[6], gene_id, gene_name])
            fout.write(gene_info + '\n')
    fin.close()
    fout.close()


def annotation_snp_pos(res_file, bed_file, p_cut=1, dis=0, ld_file=None, r2=0.2):
    """
    add the snp information for the epistasis result file; select top significant SNP pairs based on p cut value;
    select SNP pairs bases on their distance.
    :param res_file: The result file for epistasis file. The first two columns are the orders for the SNP pairs, and
    the last column is the p values.
    :param bed_file: the prefix for the plink binary file
    :param p_cut: p cut value, default value is 1.
    :param dis: the min distance between SNP pairs, the default value is -1.
    :return: 0
    """
    snp_info = {}
    order = -1
    with open(bed_file + '.bim', 'r') as fin:
        for line in fin:
            order += 1
            arr = line.split()
            snp_info[str(order)] = ' '.join(arr)
    with open(res_file, 'r') as fin, open(res_file + '.anno', 'w') as fout:
        line = fin.readline()
        arr = line.split()
        fout.write(' '.join([arr[0], 'snp0_chr', 'snp0_ID', 'snp0_cm', 'snp0_bp', 'snp0_allele1', 'snp0_allele2',
                         arr[1], 'snp1_chr', 'snp1_ID', 'snp1_cm', 'snp1_bp', 'snp1_allele1', 'snp1_allele2']))
        fout.write(' ')
        fout.write(' '.join(arr[2:]))
        fout.write('\n')
        for line in fin:
            arr = line.split()
            snp0 = snp_info[arr[0]].split()
            snp1 = snp_info[arr[1]].split()
            if float(arr[-1]) <= p_cut and (snp0[0] != snp1[0] or abs(float(snp0[3]) - float(snp1[3])) > dis):
                fout.write(' '.join([arr[0], snp_info[arr[0]], arr[1], snp_info[arr[1]]]))
                fout.write(' ')
                fout.write(' '.join(arr[2:]))
                fout.write('\n')
    if ld_file is not None:
        ld_id= {}
        with open(ld_file, 'r') as fin:
            fin.readline()
            for line in fin:
                arr = line.split()
                if float(arr[-1]) > r2:
                    ld_id[' '.join([arr[2], arr[5]])] = 1
                    ld_id[' '.join([arr[5], arr[2]])] = 1
        with open(res_file + '.anno', 'r') as fin, open(res_file + '.anno.ld', 'w') as fout:
            line = fin.readline()
            fout.write(line)
            for line in fin:
                arr = line.split()
                snp = ' '.join([arr[2], arr[9]])
                if snp not in ld_id:
                    fout.write(line)
    return 0


def annotation_snp_nearest_gene(bed_file, gene_file, max_distance=150000):
    """
    寻找SNP临近基因
    :param bed_file: plink文件
    :param gene_file: 基因信息文件，前三列是染色体号、起始位置、终止位置
    :param max_distance: SNP距离基因的最大距离
    :return:
    """
    gene_info = {}
    fin = open(gene_file)
    for line in fin:
        arr = line.split()
        try:
            gene_info[arr[0]].append(arr[:])
        except Exception as e:
            del e
            gene_info[arr[0]] = [arr[:]]
    fin.close()
    fout = open(bed_file + '.nearby_genes', 'w')
    fin = open(bed_file + '.bim')
    for line in fin:
        snp_info = line.strip()
        arr = line.split()
        snp_pos = int(arr[3])
        for genei in gene_info[arr[0]]:
            start = int(genei[1])
            end = int(genei[2])
            dist1 = snp_pos - start
            dist2 = snp_pos - end
            if dist1 > 0 and dist2 < 0:
                fout.write(snp_info + ' ' + ' '.join(genei) + ' within' + '\n')
            else:
                distance = min(abs(dist1), abs(dist2))
                if distance < max_distance:
                    fout.write(snp_info + ' ' + ' '.join(genei) + ' ' + str(distance) + '\n')
    fin.close()
    fout.close()





import numpy as np
import logging
import sys


def random_pair(num_snp, out_file='random_pair', num_pair=100000, num_each_pair=5000):
    """
    Produce random SNP pairs for additive by additive epistasis and dominance by dominance epistasis
    :param num_snp: total number of SNPs
    :param out_file: output file. default is 'random_pair'
    :param num_pair: total SNP pairs to produce. default is 10000
    :param num_each_pair: the number of SNP pairs for each sampling. default is 5000
    :return: the random SNP pairs
    """
    if num_pair > num_snp*(num_snp-1)/2:
        logging.error('num_pair must be not greater than: ' + str(num_snp*(num_snp-1)/2))
        sys.exit()
    if num_pair < num_each_pair:
        logging.error('num_pair must be greater than num_each_pair')
        sys.exit()
    res_lst = set()
    while True:
        arr = np.random.randint(num_snp, size=(num_each_pair, 2))
        arr = arr[arr[:, 0] < arr[:, 1], :]
        arr = np.array(arr, dtype=str)
        lst = map(lambda x, y: x + ' ' + y, arr[:, 0], arr[:, 1])
        res_lst.update(set(lst))
        if len(res_lst) >= num_pair:
            break
    res_lst = list(res_lst)
    res_lst = res_lst[:num_pair]
    np.savetxt(out_file, np.array(res_lst), fmt="%s", header='snp_0 snp_1', comments='')
    res_lst = np.loadtxt(out_file, dtype=np.int, skiprows=1)
    return res_lst


def random_pairAD(num_snp, out_file='random_pair', num_pair=100000, num_each_pair=5000):
    """
    Produce random SNP pairs for additive by dominance epistasis
    :param num_snp: total number of SNPs
    :param out_file: output file. default is 'random_pair'
    :param num_pair: total SNP pairs to produce. default is 10000
    :param num_each_pair: the number of SNP pairs for each sampling. default is 5000
    :return: the random SNP pairs
    """
    if num_pair > num_snp*(num_snp-1):
        logging.error('num_pair must be not greater than: ' + str(num_snp*(num_snp-1)))
        sys.exit()
    if num_pair < num_each_pair:
        logging.error('num_pair must be greater than num_each_pair')
        sys.exit()
    res_lst = set()
    while True:
        arr = np.random.randint(num_snp, size=(num_each_pair, 2))
        arr = arr[arr[:, 0] != arr[:, 1], :]
        arr = np.array(arr, dtype=str)
        lst = map(lambda x, y: x + ' ' + y, arr[:, 0], arr[:, 1])
        res_lst.update(set(lst))
        if len(res_lst) >= num_pair:
            break
    res_lst = list(res_lst)
    res_lst = res_lst[:num_pair]
    np.savetxt(out_file, np.array(res_lst), fmt="%s", header='snp_0 snp_1', comments='')
    res_lst = np.loadtxt(out_file, dtype=np.int, skiprows=1)
    return res_lst




"""
additive by additive test by SNP-BLUP model. Mouse data.

"""
import logging
import numpy as np
from glta.uvlmm import wemai_multi_gmat
from gmat.gmatrix import agmat, dgmat_as
from gmat.remma.remma_cpu import remma_add_cpu

logging.basicConfig(level=logging.INFO)

# prepare the phenotypic vector, design matrixed for fixed effects and random effects
from gmat.uvlmm import design_matrix_wemai_multi_gmat

pheno_file = '../data/mouse/pheno'
bed_file = '../data/mouse/plink'
y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)

# Calculate the genomic relationship matrix
from gmat.gmatrix.gmatrix import agmat, dgmat_as
a = agmat(bed_file, inv=False)
b = dgmat_as(bed_file, inv=False)


# Example 1: Just include additive relationship matrix
from gmat.uvlmm.uvlmm_varcom import wemai_multi_gmat
gmat_lst = [a[0]]
var_com_a = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a, bed_file, out_file='../data/mouse/remma_add_cpu_a')


# Example 2: Include additive relationship matrix and dominance relationship matrix
gmat_lst = [a[0], b[0]]
var_com_a_d = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a_d = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a_d, bed_file, out_file='../data/mouse/remma_add_cpu_a_d')


# Example 3: Include additive relationship matrix and additive by additive relationship matrix
gmat_lst = [a[0], a[0]*a[0]]
var_com_a_axa = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a_axa = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a_axa, bed_file, out_file='../data/mouse/remma_add_cpu_a_axa')


# Example 4: Include additive, dominance, additive by additive, additive by dominance,
# dominance by dominance relationship matrix
gmat_lst = [a[0], b[0], a[0]*a[0], a[0]*b[0], b[0]*b[0]]
var_com_all = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_all = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_all, bed_file, out_file='../data/mouse/remma_add_cpu_all')

print('Median values: ', np.median(res_a['p_val']), np.median(res_a_d['p_val']), np.median(res_a_axa['p_val']),
      np.median(res_all['p_val']))
print('Pearson product-moment correlation coefficients: ', np.corrcoef([res_a['p_val'], res_a_d['p_val'], res_a_axa['p_val'], res_all['p_val']]))





import pandas as pd
import numpy as np
from scipy.stats import chi2
import logging
from glta.gmat import agmat, dgmat_as
from glta.uvlmm import wemai_multi_gmat

from glta.remma.remma_cpu import remma_add_cpu
from glta.remma.remma_cpu import remma_dom_cpu

from glta.remma.remma_cpu import remma_epiAA_cpu, remma_epiAA_select_cpu, remma_epiAA_pair_cpu, \
    remma_epiAA_eff_cpu, remma_epiAA_eff_cpu_c

from glta.remma.remma_cpu import remma_epiAD_cpu, remma_epiAD_select_cpu, remma_epiAD_pair_cpu, \
    remma_epiAD_eff_cpu, remma_epiAD_eff_cpu_c

from glta.remma.remma_cpu import remma_epiDD_cpu, remma_epiDD_select_cpu, remma_epiDD_pair_cpu, \
    remma_epiDD_eff_cpu, remma_epiDD_eff_cpu_c




logging.basicConfig(level=logging.INFO)



bed_file = '../data/plink'
df = pd.read_csv("../data/pheno", header=None, sep='\s+')

y = np.array(df.iloc[:, -1:])
xmat = np.array(df.iloc[:, 2:-1])



agmat_lst = agmat(bed_file, inv=True, small_val=0.001, out_fmt='mat')
dgmat_lst = dgmat_as(bed_file, inv=True, small_val=0.001, out_fmt='mat')


gmat_lst = [agmat_lst[0], dgmat_lst[0], agmat_lst[0]*agmat_lst[0], agmat_lst[0]*dgmat_lst[0], dgmat_lst[0]*dgmat_lst[0]]

var_com = wemai_multi_gmat(y, xmat, gmat_lst, init=None, maxiter=200, cc=1.0e-8)


res_add = remma_add_cpu(y, xmat, gmat_lst, var_com, bed_file, out_file='remma_add_cpu')
print(np.median(res_add['p_val']))

res_dom = remma_dom_cpu(y, xmat, gmat_lst, var_com, bed_file, out_file='remma_dom_cpu')
print(np.median(res_dom['p_val']))

###########加加互作
res_epiAA_cpu = remma_epiAA_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, p_cut=1, out_file='remma_epiAA_cpu')
p_vec = chi2.sf(res_epiAA_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAA_select_cpu = remma_epiAA_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiAA_select_cpu')
p_vec = chi2.sf(res_epiAA_select_cpu[:, -1], 1)
print(np.median(p_vec))


snp1 = np.arange(1, 1407).reshape(-1, 1)
snp2 = np.array([0]*1406).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiAA_pair_cpu = remma_epiAA_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiAA_pair_cpu')
p_vec = chi2.sf(res_epiAA_pair_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAA_eff_cpu = remma_epiAA_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu')

res_epiAA_eff_cpu_c = remma_epiAA_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu_c')


###########加显互作
res_epiAD_cpu = remma_epiAD_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, p_cut=1, out_file='remma_epiAD_cpu')
p_vec = chi2.sf(res_epiAD_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAD_select_cpu = remma_epiAD_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiAD_select_cpu')
p_vec = chi2.sf(res_epiAD_select_cpu[:, -1], 1)
print(np.median(p_vec))


snp1 = np.arange(1, 1407).reshape(-1, 1)
snp2 = np.array([0]*1406).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiAD_pair_cpu = remma_epiAD_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiAD_pair_cpu')
p_vec = chi2.sf(res_epiAD_pair_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAD_eff_cpu = remma_epiAD_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAD_eff_cpu')

res_epiAD_eff_cpu_c = remma_epiAD_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAD_eff_cpu_c')

###########显显互作
res_epiDD_cpu = remma_epiDD_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, p_cut=1, out_file='remma_epiDD_cpu')
p_vec = chi2.sf(res_epiDD_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiDD_select_cpu = remma_epiDD_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiDD_select_cpu')
p_vec = chi2.sf(res_epiDD_select_cpu[:, -1], 1)
print(np.median(p_vec))


snp1 = np.arange(1, 1407).reshape(-1, 1)
snp2 = np.array([0]*1406).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiDD_pair_cpu = remma_epiDD_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiDD_pair_cpu')
p_vec = chi2.sf(res_epiDD_pair_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiDD_eff_cpu = remma_epiDD_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiDD_eff_cpu')

res_epiDD_eff_cpu_c = remma_epiDD_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiDD_eff_cpu_c')







var_com = [0.06289206, 0.07641075, 0.08121168]
# res_epiAA_eff_cpu = remma_epiAA_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu')

var_com = [0.06289206, 0.07641075, 0.08121168]
res_epiAA_eff_cpu_c = remma_epiAA_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu_c')

'''
res_epiAA_select_cpu = remma_epiAA_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiAA_select_cpu')


snp1 = np.arange(1407).reshape(-1, 1)
snp2 = np.array([0]*1407).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiAA_pair_cpu = remma_epiAA_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiAA_pair_cpu')


from glta.remma import remma_random_pair

num_snp = 1000
remma_random_pair(num_snp, num_pair=100000, out_file='random_pair')


'''





def ped_trace(id_file, full_ped_file, gen=1000000):
    """
    描述
    ----
    根据给定的个体号，在全系谱文件中追溯这些个体的祖先。

    参数
    ----
    id_file: 储存ID号的文件，文件可包含任意列数，但第一列必须是待读取的ID号，不能出现缺失（即“0”）
    full_ped_file: 全系谱文件，前三列依次是个体号、父号、母号，系谱中的缺失用“0”表示。
    gen: 默认值为1000000，对于每个个体最大追溯的世代数。

    返回值
    ------
    个体及追溯到的祖先个体总数，相应的系谱关系输出到以id_file为前缀的"*.trace"文件里

    """

    # 读取个体号文件
    ped_id = {}
    ped_dct = {}
    fin = open(id_file, 'r')
    for line in fin:
        arr = line.split()
        if arr[0] == '0':
            print("0 is not allowed for id")  # 如果出现“0”，程序强制终止
            exit()
        ped_id[arr[0]] = 1
        ped_dct[arr[0]] = arr[0] + '\t0\t0'  # 给个体预设默认父母号，防止个体号没有出现在全系谱文件中
    fin.close()

    # 追溯个体的祖先
    ped_newid = set()
    for i in range(gen):
        ped_traceid = {}
        ped_len1 = len(ped_id.keys())
        fin = open(full_ped_file, 'r')
        for line in fin:
            arr = line.split()
            if arr[0] in ped_id:
                if arr[1] != '0':
                    ped_traceid[arr[1]] = 1
                if arr[2] != '0':
                    ped_traceid[arr[2]] = 1
        fin.close()
        ped_newid = set(ped_traceid.keys()) - set(ped_id.keys())
        ped_id.update(ped_traceid)
        ped_len2 = len(ped_id.keys())
        if ped_len1 == ped_len2:
            break

    # 读取系谱
    fin = open(full_ped_file)
    for line in fin:
        arr = line.split()
        ped_dct[arr[0]] = arr[0] + '\t0\t0'
        ped_dct[arr[1]] = arr[1] + '\t0\t0'
        ped_dct[arr[2]] = arr[2] + '\t0\t0'
    fin.close()
    fin = open(full_ped_file)
    for line in fin:
        arr = line.split()
        ped_dct[arr[0]] = "\t".join(arr[0:3])
    fin.close()

    # 输出
    fout = open(id_file + '.trace', 'w')
    for ikey in ped_id.keys():
        if ikey in ped_newid:
            fout.write('\t'.join([ikey, '0', '0']) + '\n')
        else:
            fout.write(ped_dct[ikey] + '\n')
    fout.close()
    return len(ped_id.keys())


def ped_correct(ped_file):
    """
    描述
    ----
    修正系谱中可能出现的错误
    (1) 个体号同时出现在父亲列和母亲列，这时将计算个体号在父亲列和母亲列出现的频率，将个体号在出现频率低的列中删除（设为缺失）；
    (2) 个体号同时出现在自己的祖先中，将对应的祖先号设为缺失，避免系谱按出生年月排序时出现死循环。

    参数
    ----
    ped_file: 系谱文件，前三列分别是个体号、父号、母号

    返回值
    ----
    修正后的系谱字典，以个体号为键、父母号组成的列表为值
    生成三个以ped_file为前缀的文件，”*.error1”存储犯第一种错误的系谱，”*.error2”存储犯第二种错误的系谱，”*.correct”存储修正后的系谱
    
    """

    # 计算个体号在父列和母列出现的频率
    sire_count = {}
    dam_count = {}
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        sire_count[arr[1]] = sire_count.get(arr[1], 0) + 1
        dam_count[arr[2]] = dam_count.get(arr[2], 0) + 1
    fin.close()

    # 将出现频率低的个体号存入列表
    id_overlap = set(sire_count.keys()) & set(dam_count.keys())   # 重叠个体号
    id_overlap.discard('0')
    sire_del, dam_del = [], []
    for val in id_overlap:
        if sire_count[val] < dam_count[val]:
            sire_del.append(val)
        else:
            dam_del.append(val)

    # 为父列和母列设置默认系谱（即个体号、0、0）
    ped_dct = {}
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        ped_dct[arr[1]] = ['0', '0']
        ped_dct[arr[2]] = ['0', '0']
    fin.close()

    # 将频率低的ID在相应列中设为缺失，并将可能出现错误的系谱输出
    fout = open(ped_file + '.error1', 'w')
    fin = open(ped_file, 'r')
    print('#Both sire and dam')
    for line in fin:
        arr = line.split()
        if arr[1] in sire_del:
            fout.write('\t'.join(arr[0:3]) + '\n')
            arr[1] = '0'
        if arr[2] in dam_del:
            fout.write('\t'.join(arr[0:3]) + '\n')
            arr[2] = '0'
        ped_dct[arr[0]] = [arr[1], arr[2]]
    fout.close()
    fin.close()
    try:
        ped_dct.pop('0')
    except Exception as e:
        del e

    # 将个体号所有祖先存入字典
    anc = {}
    for id in list(ped_dct.keys()):
        anc[id] = {}
        if ped_dct[id][0] != '0':
            anc[id][ped_dct[id][0]] = 1
        if ped_dct[id][1] != '0':
            anc[id][ped_dct[id][1]] = 1
        d = 1000
        while d > 0:
            id_key = list(anc[id].keys())
            id_key_len1 = len(id_key)
            for i in id_key:
                if i in ped_dct:
                    anc[id][ped_dct[i][0]] = 0
                    anc[id][ped_dct[i][1]] = 0
            id_key = list(anc[id].keys())
            id_key_len2 = len(id_key)
            d = id_key_len2 - id_key_len1
        try:
            anc[id].pop('0')
        except Exception as e:
            del e
    
    # 当个体号同时出现在其祖先时，对应祖先设为缺失
    fout = open(ped_file + '.error2', 'w')
    print('#Both id and its ancestor')
    for i in list(anc.keys()):
        ikey_lst = list(anc[i].keys())
        if i in ikey_lst:  # 判断个体号同时是其祖先
            stri = i + '\t' + ped_dct[i][0] + '\t' + ped_dct[i][1] + '\n'  # 输出个体号系谱
            fout.write(stri)
            for j in ikey_lst:  # 祖先对应的系谱条目
                if ped_dct[j][0] == i:
                    stri = j + '\t' + ped_dct[j][0] + '\t' + ped_dct[j][1] + '\n'  # 输出个体号同时出现在祖先中的对应系谱
                    fout.write(stri)
                    ped_dct[j][0] = '0'  # 将错误祖先个体号设为缺失
                if ped_dct[j][1] == i:
                    stri = j + '\t' + ped_dct[j][0] + '\t' + ped_dct[j][1] + '\n'  # 输出个体号同时出现在祖先中的对应系谱
                    fout.write(stri)
                    ped_dct[j][1] = '0'  # 将错误祖先个体号设为缺失
    fout.close()
    
    # 输出修正后的系谱
    fout = open(ped_file + '.correct', 'w')
    for i in ped_dct.keys():
        stri = i + '\t' + ped_dct[i][0] + '\t' + ped_dct[i][1] + '\n'
        fout.write(stri)
    fout.close()
    return ped_dct


def ped_sort(ped_file):
    """
    描述
    ----
    将系谱按出生先后排序，即个体祖先的系谱条目要出现在个体之前
    
    参数
    ----
    ped_file: 系谱文件，前三列分别是个体号、父号、母号
    
    返回值
    ------
    0，生成以原系谱文件为前缀的排序后文件".sort"
    
    """
    
    # 为系谱中所有个体号设置默认系谱
    ped_dct = {}
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        ped_dct[arr[0]] = ['0', '0']
        ped_dct[arr[1]] = ['0', '0']
        ped_dct[arr[2]] = ['0', '0']
    fin.close()
    
    # 读取系谱
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        ped_dct[arr[0]] = [arr[1], arr[2]]
    fin.close()
    
    try:
        ped_dct.pop('0')
    except Exception as e:
        del e
    
    # 输出排序后系谱
    output = {'0': 1}  # 存储已输出的个体号
    ped_len = len(ped_dct.keys())
    fout = open(ped_file + '.sort', 'w')
    while ped_len > 0:
        for ikey in list(ped_dct.keys()):
            # 如果亲本均已输出，则输出此个体系谱，并将此个体号存入字典
            if ped_dct[ikey][0] in output and ped_dct[ikey][1] in output:
                stri = '\t'.join([ikey, ped_dct[ikey][0], ped_dct[ikey][1]])
                fout.write(stri + '\n')
                output[ikey] = 1
                ped_dct.pop(ikey)
        ped_len = len(ped_dct.keys())
    fout.close()
    return 0


def ped_recode(ped_file):
    """
    描述
    ----
    将系谱中的个体号编码成从1开始的整数
    
    参数
    ----
    ped_file: 系谱文件，前三列分别是个体号、父号、母号
    
    返回值
    ------
    若成功，返回整数0，并生成两个以ped_file为前缀的文件，"*.dct"为个体号与其编码对应文件，第一列为个体号，第二列为编码后的整数；
    "*.recode"为编码后的文件
    
    """
    
    # 为系谱第一列编码
    code_dct = {"0": 0}
    code_val = 0
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        if arr[0] not in code_dct:
            code_val += 1
            code_dct[arr[0]] = code_val
    fin.close()
    
    # 为系谱第二、三列编码，并输出编码后系谱
    fout = open(ped_file + '.recode', 'w')
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        if arr[1] not in code_dct:
            code_val += 1
            code_dct[arr[1]] = code_val
        if arr[2] not in code_dct:
            code_val += 1
            code_dct[arr[2]] = code_val
        stri = "{0:d}\t{1:d}\t{2:d}\n".format(code_dct[arr[0]], code_dct[arr[1]], code_dct[arr[2]])
        fout.write(stri)
    fin.close()
    fout.close()
    
    # 输出个体号与编码对应文件
    fout = open(ped_file + '.dct', 'w')
    code_dct.pop('0')
    for ikey in code_dct.keys():
        fout.write('{0:s}\t{1:d}\n'.format(ikey, code_dct[ikey]))
    fout.close()
    return 0


def ped_completeness(ped_file, gen=5, cut=0.8):
    """
    描述
    ----
    计算个体的系谱完整度，并筛选出系谱完整度高的个体。参考文献：MacCluer et al. Inbreeding and pedigree structure in
    Standardbred horses, The Journal of Heredity 74:394-399. 1983.
    
    参数
    ----
    ped_file: 系谱文件
    gen: 默认为5，计算个体系谱完整度往前追溯的世代数
    cut: 默认为0.8，系谱筛选阈值，系谱完整度低于阈值的个体将被删除
    
    返回值
    ------
    生成两个以ped_file为前缀的文件
    "*.pec": 个体系谱复杂度文件，仅包含高于阈值的个体信息，第一列为个体号，第二列为系谱完整度
    "*.prune": 质控后的系谱信息，包含高于阈值及其gen（用于计算系谱完整度的代数）代祖先个体的系谱信息
    
    """
    
    # 为系谱中所有个体号设置默认系谱
    ped_dct = {}
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        ped_dct[arr[0]] = ['0', '0']
        ped_dct[arr[1]] = ['0', '0']
        ped_dct[arr[2]] = ['0', '0']
    fin.close()
    
    # 读取系谱信息
    fin = open(ped_file, 'r')
    for line in fin:
        arr = line.split()
        ped_dct[arr[0]] = [arr[1], arr[2]]
    fin.close()
    
    # 计算个体系谱完整度
    ped_output = {}  # 储存将被输出的个体号
    fout = open(ped_file + '.pec', 'w')
    sire1, dam1 = [], []
    for ikey in list(ped_dct.keys()):
        anc_lst = []
        if ped_dct[ikey][0] == '0' or ped_dct[ikey][1] == '0':  # 如果个体亲本有一缺失，完整度即为0
            pec_val = 0.0
        else:
            sire1, dam1 = [ped_dct[ikey][0]], [ped_dct[ikey][1]]
            anc_lst.extend(sire1)
            anc_lst.extend(dam1)
            pec_sire, pec_dam = 0.5, 0.5  # 往前一代父系、母系完整度
            for val in range(2, gen + 1):  # 往前2-gen代
                sire2, dam2 = [], []
                for id in sire1:
                    if ped_dct[id][0] != '0':
                        pec_sire += 1.0/pow(2, val)  # pow(2, val) 为往前第val世代亲本数
                        sire2.append(ped_dct[id][0])
                    if ped_dct[id][1] != '0':
                        pec_sire += 1.0/pow(2, val)
                        sire2.append(ped_dct[id][1])
                for id in dam1:
                    if ped_dct[id][0] != '0':
                        pec_dam += 1.0/pow(2, val)
                        dam2.append(ped_dct[id][0])
                    if ped_dct[id][1] != '0':
                        pec_dam += 1.0/pow(2, val)
                        dam2.append(ped_dct[id][1])
                sire1, dam1 = list(sire2), list(dam2)
                anc_lst.extend(sire1)
                anc_lst.extend(dam1)
            pec_sire /= gen
            pec_dam /= gen
            pec_val = 4 * pec_sire * pec_dam / (pec_sire + pec_dam)
        if pec_val > cut:
            fout.write("{0:s}\t{1:f}\n".format(ikey, pec_val))
            ped_output[ikey] = list(ped_dct[ikey])
            for id in anc_lst:
                if id in sire1 or id in dam1:  # 往前第gen代父母设为0、0，否则相当于追溯了gen + 1代
                    ped_output[id] = ['0', '0']
                else:
                    ped_output[id] = list(ped_dct[id])
    fout.close()
    
    # 输出质控后的系谱
    fout = open(ped_file + '.prune', 'w')
    for id in ped_output:
        fout.write(id + '\t' + '\t'.join(ped_output[id]) + '\n')
    fout.close()
    
    return 0



from gmat.uvlmm.varcom.em import em_mme, em_vmat
from gmat.uvlmm.varcom.em2 import em_mme2, em_vmat2
from gmat.uvlmm.varcom.em22 import em_mme22, em_vmat22


from gmat.uvlmm.varcom.pxem import pxem_mme, pxem_vmat
from gmat.uvlmm.varcom.pxem2 import pxem_mme2, pxem_vmat2


from gmat.uvlmm.varcom.ai import ai_vmat, ai_mme


from gmat.uvlmm.varcom.weight_emai import weight_emai_vmat, weight_emai_mme

from gmat.uvlmm.varcom.weight_pxemai import weight_pxemai_vmat, weight_pxemai_mme



import math
import numpy as np
from scipy import linalg
import datetime
import pandas as pd
import logging

from .common import *
from .pre_mat import *
from .iter_mat import *


def balance_emai(y, xmat, leg_tp, kin_eigen_val, init=None, maxiter=30, cc_par=1.0e-8, cc_gra=1.0e-6, em_weight_step=0.02):
    num_record = y.shape[0]*y.shape[1]*y.shape[2]
    ran_df = len(kin_eigen_val)
    cov_dim = leg_tp.shape[1]
    var_com = np.array(init)
    var_com_update = var_com.copy()
    delta = 1000.0
    logging.info("initial variances: " + ' '.join(np.array(var_com, dtype=str)))
    cov = pre_cov_mat_eigen(cov_dim, var_com)
    if cov is None:
        logging.info('The covariances are not positive defined!')
        exit()
    cov_add, cov_per, cov_res = cov[:]
    var_ind = np.tril_indices_from(cov_add)
    a = [0] * len(var_ind[0])
    a.extend([1]*len(var_ind[0]))
    a.append(2)
    b = list(var_ind[0])
    b.extend(b)
    b.append(0)
    c = list(var_ind[1])
    c.extend(c)
    c.append(0)
    var_ind = np.array([a, b, c]).T
    kin_dct = {
        0: kin_eigen_val.reshape(len(kin_eigen_val), 1, 1),
        1: 1.0,
        2: 1.0
    }
    iter_count = 0
    cc_par_val = 1000.0
    cc_gra_val = 1000.0
    while iter_count < maxiter:
        iter_count += 1
        logging.info('***Start the iteration: {} ***'.format(iter_count))
        logging.info("fd and ai matrix")
        fd_mat, ai_mat = pre_fdai_mat_eigen_glm(y, xmat, leg_tp, kin_eigen_val, cov_add, cov_per, cov_res, var_com,
                                            var_ind, kin_dct)
        logging.info("EM matrix")
        em_mat = pre_em_mat_eigen(cov_dim, cov_add, cov_per, ran_df, var_com, num_record)
        gamma = -em_weight_step
        while gamma < 1.0:
            gamma = gamma + em_weight_step
            if gamma >= 1.0:
                gamma = 1.0
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_update = var_com + delta
            cov = pre_cov_mat_eigen(cov_dim, var_com_update)
            if cov is not None:
                logging.info('EM weight value: ' + str(gamma))
                break
        logging.info('Updated variances: ' + ' '.join(np.array(var_com_update, dtype=str)))
        cov_add, cov_per, cov_res = cov[:]
        # Convergence criteria
        cc_par_val = np.sum(pow(delta, 2)) / np.sum(pow(var_com_update, 2))
        cc_par_val = np.sqrt(cc_par_val)
        cc_gra_val = np.sqrt(np.sum(pow(fd_mat, 2))) / len(var_com)
        var_com = var_com_update.copy()
        logging.info("Change in parameters, Norm of gradient vector: {} {}".format(cc_par_val, cc_gra_val))
        if cc_par_val < cc_par and cc_gra_val < cc_gra:
            break
    if cc_par_val < cc_par and cc_gra_val < cc_gra:
        logging.info("Variances Converged")
    else:
        logging.info("Variances Not Converged")
    var_pd = {'vari': np.array(var_ind[:, 0]) + 1,
              "varij": np.array(var_ind[:, 1]) + 1,
              "varik": np.array(var_ind[:, 2]) + 1,
              "var_val": var_com}
    var_pd = pd.DataFrame(var_pd, columns=['vari', "varij", "varik", "var_val"])
    return var_pd



import numpy as np
import pandas as pd
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, hstack
from scipy.stats import chi2
import datetime
import gc
import logging
from tqdm import tqdm
from patsy import dmatrix

from gmat.process_plink.process_plink import read_plink, impute_geno
from .common import *


def unbalance_longwas_fixed(data_file, id, tpoint, trait, bed_file, kin_file, var_com, snp_lst=None,
            tfix=None, fix=None, forder=3, aorder=3, porder=3, na_method='omit',
                             prefix_outfile='unbalance_longwas_fixed'):
    """
    the longitudinal GWAS for the unbalanced data treating the SNP as the time varied fixed effect.
    :param data_file: the data file. The first row is the variate names whose first initial position is alphabetical.
    For the class variates, the first letter must be capital; for the covariates (continuous variates), the first letter
    must be lowercase.
    :param id: A class variate name which indicates the individual id column in the data file.
    :param tpoint: A covariate names which indicates the time point column in the data file.
    :param trait: A variate name which indicates the analyzed trait column in the data file.
    :param bed_file: the prefix for the plink binary file.
    :param kin_file: the file for genomic relationship matrix. This file can be produced by
    gmat.gmatrix.agmat function using agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')
    :param var_com: the estimated variance parameters by the gmat.longwas.unbalance.unbalance_varcom function.
    :param snp_lst: the snp list to test. Default is None.
    :param tfix: A class variate name for the time varied fixed effect. Default value is None. Only one time varied
    fixed effect can be included in the current version.
    :param fix: Expression for the time independent fixed effect. Default value is None. An example:
    fix = "Sex + age + Season".
    :param forder: the order of Legendre polynomials for the time varied fixed effect. The default value is 3.
    :param aorder: the order of Legendre polynomials for the additive genetic effect. The default value is 3.
    :param porder: the order of Legendre polynomials for the permanent environment effect. The default value is 3.
    :param na_method: The method to deal with missing values. The default value is 'omit'. 'omit' method will delete the
    row with missing values. 'include' method will fill the missing values with the adjacent values.
    :param prefix_outfile: the prefix for the output file. Default is 'unbalance_longwas_fixed'.
    :return: A pandas data frame for the test result.
    """
    logging.info('################################')
    logging.info('###Prepare the related matrix###')
    logging.info('################################')
    if var_com.shape[0] != aorder*(aorder+1)/2 + aorder + 1 + porder*(porder+1)/2 + porder + 1 + 1:
        logging.info('ERROR: Variances do not match the data, please check')
        exit()
    logging.info('***Read the data file***')
    logging.info('Data file: ' + data_file)
    data_df = pd.read_csv(data_file, sep='\s+', header=0)
    logging.info('NA method: ' + na_method)
    if na_method == 'omit':
        data_df = data_df.dropna()
    elif na_method == 'include':
        data_df = data_df.fillna(method='ffill')
        data_df = data_df.fillna(method='bfill')
    else:
        logging.info('na_method does not exist: ' + na_method)
        exit()
    col_names = data_df.columns
    logging.info('The column names of data file: ' + ' '.join(list(col_names)))
    logging.info('Note: Variates beginning with a capital letter is converted into factors.')
    class_vec = []
    for val in col_names:
        if not val[0].isalpha():
            logging.info("The first character of columns names must be alphabet!")
            exit()
        if val[0] == val.capitalize()[0]:
            class_vec.append(val)
            data_df[val] = data_df[val].astype('str')
        else:
            try:
                data_df[val] = data_df[val].astype('float')
            except Exception as e:
                logging.info(e)
                logging.info(val + " may contain string, please check!")
                exit()
    logging.info('Individual column: ' + id)
    if id not in col_names:
        logging.info(id + ' is not in the data file, please check!')
        exit()
    if id not in class_vec:
        logging.info('The initial letter of {} should be capital'.format(id))
        exit()
    id_order = []
    id_arr = list(data_df[id])
    id_order.append(id_arr[0])
    for i in range(1, len(id_arr)):
        if id_arr[i] != id_arr[i - 1]:
            id_order.append(id_arr[i])
    id_in_data = set(data_df[id])
    if len(id_in_data) - len(id_order) != 0:
        logging.info('The data is not sored by individual ID!')
        exit()
    logging.info('Time points column: ' + tpoint)
    if tpoint not in col_names:
        logging.info(tpoint + ' is not in the data file, please check!')
        exit()
    if tpoint in class_vec:
        logging.info('The initial letter of {} should be lowercase'.format(tpoint))
        exit()
    logging.info('Trait column: ' + trait)
    if trait not in col_names:
        logging.info(trait + ' is not in the data file, please check!')
        exit()
    if trait in class_vec:
        logging.info('The initial letter of {} should be lowercase'.format(trait))
        exit()
    logging.info('Code factor variables of the data file: ' + ' '.join(list(class_vec)))
    code_val = {}
    code_dct = dct_2D()
    for val in class_vec:
        code_val[val] = 0
        temp = []
        for i in range(data_df.shape[0]):
            if data_df[val][i] not in code_dct[val]:
                code_val[val] += 1
                code_dct[val][data_df[val][i]] = str(code_val[val])
            temp.append(code_dct[val][data_df[val][i]])
        data_df[val] = np.array(temp)
    for val in class_vec:
        data_df[val] = data_df[val].astype('int')
    logging.info('***Build the design matrix for fixed effect***')
    logging.info('Time dependent fixed effect: ' + str(tfix))
    leg_fix = leg(data_df[tpoint], forder)
    if tfix == None:
        xmat_t = np.concatenate(leg_fix, axis=1)
        xmat_t = csr_matrix(xmat_t)
    else:
        if tfix not in class_vec:
            logging.info(tfix + ' is not the class variate')
            exit()
        row = np.array(range(data_df.shape[0]))
        col = np.array(data_df[tfix]) - 1
        val = np.array([1.0] * data_df.shape[0])
        tfix_mat = csr_matrix((val, (row, col)))
        xmat_t = []
        for i in range(len(leg_fix)):
            xmat_t.append(tfix_mat.multiply(leg_fix[i]))
        xmat_t = hstack(xmat_t)
        del row, col, val
        gc.collect()
    logging.info('Time independent fix effect: ' + str(fix))
    xmat_nt = None
    if fix == None:
        xmat_nt = None
    else:
        try:
            fix_exp = ''
            vec = fix.split('+')
            for i in vec:
                val = i.strip()
                if val in class_vec:
                    fix_exp += 'C(' + val + ')'
                else:
                    fix_exp += val
            xmat_nt = dmatrix(fix_exp, data_df)
            logging.info('The expression for fixed effect: ' + fix_exp)
        except Exception as e:
            logging.info(e + ': Check the fix effect expression.')
            exit()
        xmat_nt = csr_matrix(xmat_nt[:, 1:])
    xmat = hstack([xmat_t, xmat_nt])
    xmat = xmat.toarray()
    max_id = max(data_df[id]) + 1
    tmin = min(data_df[tpoint])
    tmax = max(data_df[tpoint])
    leg_lst = []  # legendre polynomials for time dependent fixed SNP effects, save for each individuals
    for i in range(1, max_id):
        leg_lst.append(leg_mt(data_df[data_df[id] == i][tpoint], tmax, tmin, forder))
    tpoint_vec = sorted(set(data_df[tpoint]))
    leg_tpoint_mat = leg_mt(np.array(tpoint_vec), tmax, tmin, forder)
    leg_tpoint_accum = np.sum(leg_tpoint_mat, axis=0)
    logging.info('***Read the kinship matrix***')
    logging.info('Kinship file: ' + kin_file)
    with open(kin_file) as fin:
        row = []
        col = []
        kin = []
        id_in_kin = {}
        for line in fin:
            arr = line.split()
            id_in_kin[arr[0]] = 1
            id_in_kin[arr[1]] = 1
            if arr[0] not in code_dct[id]:
                logging.info(arr[0] + ' is not in the kinship inversion file!')
                exit()
            if arr[1] not in code_dct[id]:
                logging.info(arr[1], 'is not in the kinship inversion file!')
                exit()
            row.append(int(code_dct[id][arr[0]]))
            col.append(int(code_dct[id][arr[1]]))
            kin.append(float(arr[2]))
    id_not_in_kin = list(set(code_dct[id].keys()) - set(id_in_kin.keys()))
    if len(id_not_in_kin) != 0:
        logging.info('The ID: {} in the data file is not in the kinship file!'.format(' '.join(id_not_in_kin)))
        exit()
    kin = csr_matrix((np.array(kin), (np.array(row) - 1, np.array(col) - 1))).toarray()
    kin = np.add(kin, kin.T)
    kin[np.diag_indices_from(kin)] = 0.5 * np.diag(kin)
    del row, col
    gc.collect()
    logging.info('***Build the dedign matrix for random effect***')
    logging.info('Legendre order for additive effects: ' + str(aorder))
    leg_add = leg(data_df[tpoint], aorder)
    row = np.array(range(data_df.shape[0]))
    col = np.array(data_df[id]) - 1
    val = np.array([1.0] * data_df.shape[0])
    add_mat = csr_matrix((val, (row, col)), shape=(data_df.shape[0], kin.shape[0]))
    zmat_add = []
    for i in range(len(leg_add)):
        zmat_add.append(add_mat.multiply(leg_add[i]))
    logging.info('Legendre order for permanent environmental effect: ' + str(porder))
    leg_per = leg(data_df[tpoint], porder)
    per_mat = csr_matrix((val, (row, col)))
    zmat_per = []
    for i in range(len(leg_per)):
        zmat_per.append((per_mat.multiply(leg_per[i])))
    del row, col, val
    gc.collect()
    zmat = [zmat_add, zmat_per]
    y = data_df[trait].values.reshape(data_df.shape[0], 1)
    # kin_inv = [kin_inv, sparse.eye(max(data_df[id]), format="csr")]
    logging.info('***Prepare the merged Z matrix***')
    eff_ind = [[0, xmat.shape[1]]]  # the index for all effects [start end]
    zmat_con_lst = []  # combined random matrix
    for i in range(len(zmat)):
        temp = [eff_ind[i][-1]]
        zmat_con_lst.append(hstack(zmat[i]))
        for j in range(len(zmat[i])):
            temp.append(temp[-1] + zmat[i][j].shape[1])
        eff_ind.append(temp)
    logging.info('***Calculate the phenotypic (co)variance***')
    add_cov = var_com.loc[var_com.loc[:, 'vari']==1, :]
    row = np.array(add_cov['varij']) - 1
    col = np.array(add_cov['varik']) - 1
    val = add_cov['var_val']
    add_cov = csr_matrix((val, (row, col))).toarray()
    add_cov = add_cov + np.tril(add_cov, k=-1).T
    per_cov = var_com.loc[var_com.loc[:, 'vari']==2, :]
    row = np.array(per_cov['varij']) - 1
    col = np.array(per_cov['varik']) - 1
    val = per_cov['var_val']
    per_cov = csr_matrix((val, (row, col))).toarray()
    per_cov = per_cov + np.tril(per_cov, k=-1).T
    res_var = np.array(var_com['var_val'])[-1]
    vmat = zmat_con_lst[0].dot((zmat_con_lst[0].dot(np.kron(add_cov, kin))).T)
    one_id = sparse.eye(zmat_con_lst[1].shape[1]/per_cov.shape[0])
    vmat = vmat + zmat_con_lst[1].dot((zmat_con_lst[1].dot(sparse.kron(per_cov, one_id))).T)
    vmat_diag = np.diag(vmat) + res_var
    np.fill_diagonal(vmat, vmat_diag)
    vmat = linalg.inv(vmat)
    logging.info('***Read the snp data***')
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        logging.info('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    num_id = snp_mat.shape[0]
    num_snp = snp_mat.shape[1]
    logging.info("There are {:d} individuals and {:d} SNPs.".format(num_id, num_snp))
    fam_df = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id_geno = list(np.array(fam_df.iloc[:, 1], dtype=str))
    id_order_index = []
    for i in id_order:
        id_order_index.append(id_geno.index(i))
    if snp_lst is None:
        snp_lst = range(num_snp)
    snp_lst = list(snp_lst)
    if min(snp_lst) < 0 or max(snp_lst) >= num_snp:
        logging.info('The value in the snp list should be >= {} and < {}', 0, num_snp)
        exit()
    snp_mat = snp_mat[id_order_index, :]
    snp_mat = snp_mat[:, snp_lst]
    logging.info('#####################################################################')
    logging.info('###Start the fixed regression longitudinal GWAS for unbalance data###')
    logging.info('#####################################################################')
    chi_df = leg_lst[0].shape[1]
    eff_vec = []
    chi_vec = []
    p_vec = []
    p_min_vec = []
    p_accum_vec = []
    for i in tqdm(range(snp_mat.shape[1])):
        snp_fix = list(map(lambda x, y: x*y, leg_lst, list(snp_mat[:, i])))
        snp_fix = np.concatenate(snp_fix, axis=0)
        snp_fix = np.concatenate((xmat, snp_fix), axis=1)
        xv = np.dot(snp_fix.T, vmat)
        xvx = np.dot(xv, snp_fix)
        xvx = np.linalg.inv(xvx)
        xvy = np.dot(xv, y)
        b = np.dot(xvx, xvy)
        eff = b[-chi_df:, -1]
        eff_var = xvx[-chi_df:, -chi_df:]
        chi_val = np.sum(np.dot(np.dot(eff.T, np.linalg.inv(eff_var)), eff))
        p_val = chi2.sf(chi_val, chi_df)
        eff_vec.append(b[-chi_df:, -1])
        chi_vec.append(chi_val)
        p_vec.append(p_val)
        p_tpoint_vec = []
        for k in range(leg_tpoint_mat.shape[0]):
            eff_tpoint = np.sum(np.dot(leg_tpoint_mat[k, :], eff))
            eff_var_tpoint = np.sum(np.dot(leg_tpoint_mat[k, :], np.dot(eff_var, leg_tpoint_mat[k, :])))
            chi_tpoint = eff_tpoint*eff_tpoint/eff_var_tpoint
            p_tpoint = chi2.sf(chi_tpoint, 1)
            p_tpoint_vec.append(p_tpoint)
        p_min_vec.append(min(p_tpoint_vec))
        eff_accum = np.sum(np.dot(leg_tpoint_accum, eff))
        eff_var_accum = np.sum(np.dot(leg_tpoint_accum, np.dot(eff_var, leg_tpoint_accum)))
        chi_accum = eff_accum*eff_accum/eff_var_accum
        p_accum = chi2.sf(chi_accum, 1)
        p_accum_vec.append(p_accum)
    logging.info('Finish association analysis')
    logging.info('***Output***')
    snp_info_file = bed_file + '.bim'
    snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
    res_df = snp_info.iloc[snp_lst, [0, 1, 3, 4, 5]]
    res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
    res_df.loc[:, 'order'] = snp_lst
    res_df = res_df.iloc[:, [5, 0, 1, 2, 3, 4]]
    eff_vec = np.array(eff_vec)
    for i in range(eff_vec.shape[1]):
        col_ind = 'eff' + str(i)
        res_df.loc[:, col_ind] = eff_vec[:, i]
    res_df.loc[:, 'chi_val'] = chi_vec
    res_df.loc[:, 'p_val'] = p_vec
    res_df.loc[:, 'p_min'] = p_min_vec
    res_df.loc[:, 'p_accum'] = p_accum_vec
    out_file = prefix_outfile + '.res'
    res_df.to_csv(out_file, sep=' ', index=False)
    return res_df



# ext_build.py

from os.path import dirname, join, realpath
import cffi

ffi = cffi.FFI() #生成cffi实例

ffi.cdef("""int remma_epiAA_maf_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
        long long len_snp_lst_0, double *pymat, long long *freq, double *eff_cut, char* out_file);""")  # 函数声明



ffi.cdef("""int remma_epiAD_maf_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
        long long len_snp_lst_0, double *pymat, long long *freqA, long long *freqD, double *eff_cut, char* out_file);""")  # 函数声明

ffi.cdef("""int remma_epiDD_maf_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
        long long len_snp_lst_0, double *pymat, long long *freq, double *eff_cut, char* out_file);""")  # 函数声明





ffi.cdef("""int remma_epiAA_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
    long long len_snp_lst_0, double *pymat, double eff_cut, char* out_file);""")  # 函数声明


ffi.cdef("""int remma_epiAD_eff_cpu(char *bed_file, long long num_id, long long num_snp, 
     long long *snp_lst_0, long long len_snp_lst_0, double *pymat, double eff_cut, char* out_file);""")  # 函数声明

ffi.cdef("""int remma_epiDD_eff_cpu(char *bed_file, long long num_id, long long num_snp, 
     long long *snp_lst_0, long long len_snp_lst_0, double *pymat, double eff_cut, char* out_file);""")  # 函数声明



ffi.set_source('_cremma_epi_eff_cpu', """
    int remma_epiAA_maf_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
        long long len_snp_lst_0, double *pymat, long long *freq, double *eff_cut, char* out_file);
    int remma_epiAD_maf_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
        long long len_snp_lst_0, double *pymat, long long *freqA, long long *freqD, double *eff_cut, char* out_file);
    int remma_epiDD_maf_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
        long long len_snp_lst_0, double *pymat, long long *freq, double *eff_cut, char* out_file);
    
    int remma_epiAA_eff_cpu(char *bed_file, long long num_id, long long num_snp, long long *snp_lst_0, 
        long long len_snp_lst_0, double *pymat, double eff_cut, char* out_file);
    int remma_epiAD_eff_cpu(char *bed_file, long long num_id, long long num_snp, 
        long long *snp_lst_0, long long len_snp_lst_0, double *pymat, double eff_cut, char* out_file);
    int remma_epiDD_eff_cpu(char *bed_file, long long num_id, long long num_snp, 
        long long *snp_lst_0, long long len_snp_lst_0, double *pymat, double eff_cut, char* out_file);
    
""",
sources=[join("gmat/remma", "_remma_epi_eff_cpu.c")],
extra_compile_args = ['-fopenmp', '-lm'],
extra_link_args = ['-fopenmp', '-lm'],
)


if __name__ == '__main__':
    ffi.compile(verbose=True)




from gmat.remma.remma_epiAD.remma_epiAD import remma_epiAD, remma_epiAD_parallel
from gmat.remma.remma_epiAD.remma_epiAD_pair import remma_epiAD_pair
from gmat.remma.remma_epiAD.remma_epiAD_eff import remma_epiAD_eff, remma_epiAD_eff_parallel
from gmat.remma.remma_epiAD.remma_epiAD_approx import remma_epiAD_approx, remma_epiAD_approx_parallel
from gmat.remma.remma_epiAD.remma_epiAD_maf_approx import remma_epiAD_maf_approx, remma_epiAD_maf_approx_parallel


from gmat.remma.remma_epiAD.remma_epiAD import _remma_epiAD, _remma_epiAD_parallel
from gmat.remma.remma_epiAD.remma_epiAD_pair import _remma_epiAD_pair
from gmat.remma.remma_epiAD.remma_epiAD_eff import _remma_epiAD_eff, _remma_epiAD_eff_parallel



from gmat.remma.remma_epiDD.remma_epiDD import remma_epiDD, remma_epiDD_parallel
from gmat.remma.remma_epiDD.remma_epiDD_pair import remma_epiDD_pair
from gmat.remma.remma_epiDD.remma_epiDD_eff import remma_epiDD_eff, remma_epiDD_eff_parallel
from gmat.remma.remma_epiDD.remma_epiDD_approx import remma_epiDD_approx, remma_epiDD_approx_parallel
from gmat.remma.remma_epiDD.remma_epiDD_maf_approx import remma_epiDD_maf_approx, remma_epiDD_maf_approx_parallel


from gmat.remma.remma_epiDD.remma_epiDD import _remma_epiDD, _remma_epiDD_parallel
from gmat.remma.remma_epiDD.remma_epiDD_pair import _remma_epiDD_pair
from gmat.remma.remma_epiDD.remma_epiDD_eff import _remma_epiDD_eff, _remma_epiDD_eff_parallel




### Estimate the variances
import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.longwas.balance import balance_varcom
import logging

logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse_long/plink'
agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')

data_file = '../data/mouse_long/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = '../data/mouse_long/plink.agrm2'
prefix_outfile = '../data/mouse_long/balance_varcom'
res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
print(res_var)

### longitudinal GWAS by fixed method

import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.longwas.balance import balance_varcom
from gmat.longwas.balance import balance_longwas_fixed
from gmat.longwas.balance import balance_longwas_fixed_permutation
import logging

logging.basicConfig(level=logging.INFO)

logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse_long/plink'
# agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')

data_file = '../data/mouse_long/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = '../data/mouse_long/plink.agrm2'
prefix_outfile = '../data/mouse_long/balance_varcom'
# res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
var_com = pd.read_csv('../data/mouse_long/balance_varcom.var', header=0, sep='\s+')
prefix_outfile = '../data/mouse_long/balance_longwas_fixed'
res_fixed = balance_longwas_fixed(data_file, id, tpoint, trait, kin_file, bed_file, var_com, maxiter=0, snp_lst=None,
                                 prefix_outfile=prefix_outfile)
res_fixed_perm = balance_longwas_fixed_permutation(data_file, id, tpoint, trait, kin_file, bed_file, var_com, maxiter=0, snp_lst=None,
                                 prefix_outfile=prefix_outfile)


### longitudinal GWAS by trans method

import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.longwas.balance import balance_varcom
from gmat.longwas.balance import balance_longwas_trans
from gmat.longwas.balance import balance_longwas_trans_permutation
import logging

logging.basicConfig(level=logging.INFO)

logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse_long/plink'
# agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')

data_file = '../data/mouse_long/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = '../data/mouse_long/plink.agrm2'
prefix_outfile = '../data/mouse_long/balance_varcom'
# res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
var_com = pd.read_csv('../data/mouse_long/balance_varcom.var', header=0, sep='\s+')
prefix_outfile = '../data/mouse_long/balance_longwas_trans'
res_trans = balance_longwas_trans(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=None,
                                 prefix_outfile=prefix_outfile)
res_trans_perm = balance_longwas_trans_permutation(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=None,
                                 prefix_outfile=prefix_outfile)


### balance_longwas_trans + balance_longwas_fixed
import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.longwas.balance import balance_varcom
from gmat.longwas.balance import balance_longwas_fixed, balance_longwas_trans
import logging

logging.basicConfig(level=logging.INFO)

logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse_long/plink'
agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')

data_file = '../data/mouse_long/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = '../data/mouse_long/plink.agrm2'
prefix_outfile = '../data/mouse_long/balance_varcom'
res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
var_com = pd.read_csv('../data/mouse_long/balance_varcom.var', header=0, sep='\s+')
prefix_outfile = '../data/mouse_long/balance_longwas_trans'
res_trans = balance_longwas_trans(data_file, id, tpoint, trait, kin_file, bed_file, var_com,
                                  prefix_outfile=prefix_outfile)

res_trans_sel = res_trans[res_trans['p_val'] < 0.01]
snp_lst = np.array(res_trans_sel['order'])
prefix_outfile = '../data/mouse_long/balance_longwas_fixed_sel'
res_fixed_sel = balance_longwas_fixed(data_file, id, tpoint, trait, kin_file, bed_file, var_com, maxiter=0, snp_lst=snp_lst,
                                 prefix_outfile=prefix_outfile)



import os
import logging
import pandas as pd
import numpy as np
import gc
from gmat.remma.random_pair import random_pairAD
from gmat.remma.remma_epiAD.remma_epiAD_pair import remma_epiAD_pair
from gmat.remma.remma_epiAD.remma_epiAD_maf_eff import remma_epiAD_maf_eff, remma_epiAD_maf_eff_parallel
from gmat.process_plink.process_plink import read_plink, impute_geno


def remma_epiAD_maf_approx(pheno_file, bed_file, gmat_lst, var_com, p_cut=1.0e-5, num_random_pair=100000, out_file='epiAD_maf_approx'):
    """
    additive by additive epistasis test by random SNP-BLUP model based on approximate test
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: a list of genomic relationship matrixes.
    :param var_com: Estimated variances
    :param p_cut: put cut value. default value is 1.0e-5.
    :param num_random_pair: the number of random pairs to estimate the approximate variances of estimated effects. Default value is 100,000
    :param out_file: The prefix for output file. Default value is 'epiAA_maf_approx'
    :return: 0
    """
    logging.info("\n\n#####Randomly select {:d} pairs, and test these SNP pairs#####".format(num_random_pair))
    snp_df = pd.read_csv(bed_file + '.bim', header=None, sep='\s+')
    num_snp = snp_df.shape[0]  # the number of snp
    random_pairAD(num_snp, out_file=out_file + '.random_pairAD', num_pair=num_random_pair)
    remma_epiAD_pair(pheno_file, bed_file, gmat_lst, var_com, snp_pair_file=out_file + '.random_pairAD', p_cut=1,
                     out_file=out_file + '.random')
    os.remove(out_file + '.random_pairAD')
    logging.info("\n\n#####Calcualte the approximate denominator for Wald chi-square test#####")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freqA = np.sum(snp_mat, axis=0) / (2 * num_id)
    freqA[freqA > 0.5] = 1 - freqA[freqA > 0.5]  # maf
    np.savetxt(out_file + '.maf', freqA)
    freqA = np.array(freqA*20, dtype=np.longlong)
    freqA_dct = dict(zip(np.array(list(range(len(freqA))), dtype=str), freqA))
    freqD = np.sum(np.absolute(snp_mat-1.0) < 0.001, axis=0) / num_id
    del snp_mat
    gc.collect()
    freqD[freqD > 0.5] = 1 - freqD[freqD > 0.5]
    np.savetxt(out_file + '.heter', freqD)
    freqD = np.array(freqD * 20, dtype=np.longlong)
    freqD_dct = dict(zip(np.array(list(range(len(freqD))), dtype=str), freqD))
    median_dct = {}
    median_count_dct = {}
    with open(out_file + '.random') as fin:
        fin.readline()
        for line in fin:
            arr = line.split()
            key_val = ' '.join([str(freqA_dct[arr[0]]), str(freqD_dct[arr[1]])])
            median_count_dct[key_val] = median_count_dct.get(key_val, 0) + 1
            median_dct[key_val] = median_dct.get(key_val, 0.0) + float(arr[-3])
    all_median = 0
    all_count = 0
    for val in median_count_dct:
        all_median += median_dct[val]
        all_count += median_count_dct[val]
        median_dct[val] = median_dct[val] / median_count_dct[val]
    all_median = all_median / all_count
    freq_deno = np.ones(111)
    with open(out_file + '.freq_denominator', 'w') as fout:
        for key1 in set(freqA):
            for key2 in set(freqD):
                key_val = ' '.join([str(key1), str(key2)])
                if key_val not in median_dct:
                    median_dct[key_val] = all_median
                fout.write(key_val + ' ' + str(median_dct[key_val]) + '\n')
                freq_deno[key1 * 10 + key2] = median_dct[key_val]
    logging.info("\n\n#####Screen the epistatic effects and select top SNP pairs based on approximate test#####")
    remma_epiAD_maf_eff(pheno_file, bed_file, gmat_lst, var_com, freqA=freqA, freqD=freqD, freq_deno=freq_deno, p_cut=p_cut,
                    out_file=out_file + '.approx_p')
    logging.info("\n\n#####Calculate exact p values for top SNP pairs#####")
    remma_epiAD_pair(pheno_file, bed_file, gmat_lst, var_com, snp_pair_file=out_file + '.approx_p', p_cut=1,
                     out_file=out_file + '.exact_p')
    logging.info("\n\n#####Merge the results#####")
    p_dct = {}
    with open(out_file + '.approx_p', 'r') as fin:
        for line in fin:
            arr = line.split()
            p_dct[' '.join(arr[:2])] = arr[-1]
    with open(out_file + '.exact_p', 'r') as fin, open(out_file, 'w') as fout:
        for line in fin:
            arr = line.split()
            arr.insert(-1, p_dct[' '.join(arr[:2])])
            fout.write(' '.join(arr) + '\n')
    os.remove(out_file + '.approx_p')
    os.remove(out_file + '.exact_p')
    return 0



def remma_epiAD_maf_approx_parallel(pheno_file, bed_file, gmat_lst, var_com, parallel, p_cut=1.0e-5, num_random_pair=100000, out_file='epiAD_maf_approx_parallel'):
    """
    additive by additive epistasis test by random SNP-BLUP model based on approximate test
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: a list of genomic relationship matrixes.
    :param var_com: Estimated variances
    :param parallel: A list containing two integers. The first integer is the number of parts to parallel. The second
    integer is the part to run. For example, parallel = [3, 1], parallel = [3, 2] and parallel = [3, 3] mean to divide
    :param p_cut: put cut value. default value is 1.0e-5.
    :param num_random_pair: the number of random pairs to estimate the approximate variances of estimated effects. Default value is 100,000
    :param out_file: The prefix for output file. Default value is 'epiAA_approx'
    :return: 0
    """
    logging.info("\n\n#####Randomly select {:d} pairs, and test these SNP pairs#####".format(num_random_pair))
    bim_df = pd.read_csv(bed_file + '.bim', header=None)
    num_snp = bim_df.shape[0]
    random_pairAD(num_snp, out_file=out_file + '.random_pair.' + str(parallel[1]), num_pair=num_random_pair)
    remma_epiAD_pair(pheno_file, bed_file, gmat_lst, var_com, snp_pair_file=out_file + '.random_pair.' + str(parallel[1]), p_cut=1,
                     out_file=out_file + '.random.' + str(parallel[1]))
    os.remove(out_file + '.random_pair.' + str(parallel[1]))
    logging.info("\n\n#####Calcualte the approximate denominator for Wald chi-square test#####")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freqA = np.sum(snp_mat, axis=0) / (2 * num_id)
    freqA[freqA > 0.5] = 1 - freqA[freqA > 0.5]  # maf
    np.savetxt(out_file + '.maf.' + str(parallel[1]), freqA)
    freqA = np.array(freqA * 20, dtype=np.longlong)
    freqA_dct = dict(zip(np.array(list(range(len(freqA))), dtype=str), freqA))
    freqD = np.sum(np.absolute(snp_mat - 1.0) < 0.001, axis=0) / num_id
    del snp_mat
    gc.collect()
    freqD[freqD > 0.5] = 1 - freqD[freqD > 0.5]
    np.savetxt(out_file + '.heter.' + str(parallel[1]), freqD)
    freqD = np.array(freqD * 20, dtype=np.longlong)
    freqD_dct = dict(zip(np.array(list(range(len(freqD))), dtype=str), freqD))
    median_dct = {}
    median_count_dct = {}
    with open(out_file + '.random.' + str(parallel[1])) as fin:
        fin.readline()
        for line in fin:
            arr = line.split()
            key_val = ' '.join([str(freqA_dct[arr[0]]), str(freqD_dct[arr[1]])])
            median_count_dct[key_val] = median_count_dct.get(key_val, 0) + 1
            median_dct[key_val] = median_dct.get(key_val, 0.0) + float(arr[-3])
    all_median = 0
    all_count = 0
    for val in median_count_dct:
        all_median += median_dct[val]
        all_count += median_count_dct[val]
        median_dct[val] = median_dct[val] / median_count_dct[val]
    all_median = all_median / all_count
    freq_deno = np.ones(111)
    with open(out_file + '.freq_denominator.' + str(parallel[1]), 'w') as fout:
        for key1 in set(freqA):
            for key2 in set(freqD):
                key_val = ' '.join([str(key1), str(key2)])
                if key_val not in median_dct:
                    median_dct[key_val] = all_median
                fout.write(key_val + ' ' + str(median_dct[key_val]) + '\n')
                freq_deno[key1 * 10 + key2] = median_dct[key_val]
    logging.info("\n\n#####Screen the epistatic effects and select top SNP pairs based on approximate test#####")
    remma_epiAD_maf_eff_parallel(pheno_file, bed_file, gmat_lst, var_com, parallel=parallel,
                                 freqA=freqA, freqD=freqD, freq_deno=freq_deno, p_cut=p_cut, out_file=out_file + '.approx_p')
    logging.info("\n\n#####Calculate exact p values for top SNP pairs#####")
    remma_epiAD_pair(pheno_file, bed_file, gmat_lst, var_com, snp_pair_file=out_file + '.approx_p.' + str(parallel[1]), p_cut=1,
                     out_file=out_file + '.exact_p.' + str(parallel[1]))
    logging.info("\n\n#####Merge the results#####")
    p_dct = {}
    with open(out_file + '.approx_p.' + str(parallel[1]), 'r') as fin:
        for line in fin:
            arr = line.split()
            p_dct[' '.join(arr[:2])] = arr[-1]
    with open(out_file + '.exact_p.' + str(parallel[1]), 'r') as fin, open(out_file + '.' + str(parallel[1]), 'w') as fout:
        for line in fin:
            arr = line.split()
            arr.insert(-1, p_dct[' '.join(arr[:2])])
            fout.write(' '.join(arr) + '\n')
    os.remove(out_file + '.approx_p.' + str(parallel[1]))
    os.remove(out_file + '.exact_p.' + str(parallel[1]))
    return 0



from gmat.uvlmm.uvlmm_varcom import wemai_multi_gmat, _wemai_multi_gmat, wemai_multi_gmat_pred
from gmat.uvlmm.uvlmm_varcom_eigen import *
from gmat.uvlmm.uvlmm_gwas import *
from gmat.uvlmm.design_matrix import design_matrix_wemai_multi_gmat, design_matrix_wemai_multi_gmat_pred



import gc
import numpy as np
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, isspmatrix, block_diag, lil_matrix, hstack, vstack
from scipy.sparse.linalg import inv
from .common import *


# inversion of covariance matrix
def pre_covi_mat(cov_dim, var_com):
    end = 0
    covi_mat = []  # covariance matrix list
    if var_com[-1] <= 0:
        return None
    for i in range(len(cov_dim)):
        start = end
        end = int(cov_dim[i] * (cov_dim[i] - 1) / 2) + cov_dim[i] + start
        covar = np.zeros((cov_dim[i], cov_dim[i]))
        covar[np.tril_indices(cov_dim[i])] = var_com[start:end]
        covar = np.add(covar, np.tril(covar, -1).T)
        try:
            linalg.cholesky(covar)
        except:
            return None
        covar = linalg.inv(covar)
        covi_mat.append(covar)
    
    return covi_mat


def pre_mat(y, xmat, eff_ind, zmat_con, cov_dim, var_com, covi_mat, \
            cmat_pure, kin, rhs_pure):
    # Coefficient Matrix
    cmat = (cmat_pure.multiply(1.0 / var_com[-1])).toarray()
    for i in range(len(cov_dim)):
        if isspmatrix(kin[i]):
            temp = sparse.kron(covi_mat[i], kin[i])
            temp = temp.toarray()
        else:
            temp = linalg.kron(covi_mat[i], kin[i])
        
        cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
        eff_ind[i + 1][0]:eff_ind[i + 1][-1]] = \
            np.add(cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
                   eff_ind[i + 1][0]:eff_ind[i + 1][-1]], temp)
    
    # right hand
    rhs_mat = np.divide(rhs_pure, var_com[-1])
    # all effects
    cmat = linalg.inv(cmat)
    eff = np.dot(cmat, rhs_mat)
    e = y - xmat.dot(eff[:eff_ind[0][1], :]) - zmat_con.dot(
        eff[eff_ind[0][1]:, :])
    
    return cmat, eff, e


def pre_sparse_mat(y, xmat, eff_ind, zmat_con, cov_dim, var_com, covi_mat, \
            cmat_pure, kin, rhs_pure):
    # Coefficient Matrix
    cmat = cmat_pure.multiply(1.0 / var_com[-1])
    '''
    for i in range(len(cov_dim)):
        temp = sparse.kron(covi_mat[i], kin[i])
        temp = temp.tolil()
        cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
        eff_ind[i + 1][0]:eff_ind[i + 1][-1]] = \
            np.add(cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
                   eff_ind[i + 1][0]:eff_ind[i + 1][-1]], temp)
    '''
    temp_vec = []
    for i in range(len(cov_dim)):
        temp = sparse.kron(covi_mat[i], kin[i])
        temp_vec.append(np.add(cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], eff_ind[i + 1][0]:eff_ind[i + 1][-1]], temp))
    del temp
    gc.collect()
    cmat = vstack([cmat[:eff_ind[0][1], :],
     hstack([cmat[eff_ind[1][0]:eff_ind[1][-1], :eff_ind[1][0]], temp_vec[0],
             cmat[eff_ind[1][0]:eff_ind[1][-1], eff_ind[1][-1]:]]),
        hstack([cmat[eff_ind[2][0]:, :eff_ind[2][0]], temp_vec[1]])])
    del temp_vec
    gc.collect()
    # right hand
    rhs_mat = np.divide(rhs_pure, var_com[-1])
    
    # all effects
    # cmat = cmat.tocsc()
    # cmat = inv(cmat)
    # cmat = cmat.tocsr()
    # eff = cmat.dot(rhs_mat)
    # e = y - xmat.dot(eff[:eff_ind[0][1], :]) - zmat_con.dot(
    #     eff[eff_ind[0][1]:, :])
    
    return cmat, rhs_mat



def pre_mat_x(y, xmat, eff_ind, zmat_con, cov_dim, var_com, covi_mat, cmat_pure, kin, rhs_pure):
    # Coefficient Matrix
    cmat = (cmat_pure.multiply(1.0 / var_com[-1])).toarray()
    for i in range(len(cov_dim)):
        if isspmatrix(kin[i]):
            temp = sparse.kron(kin[i], covi_mat[i])
            temp = temp.toarray()
        else:
            temp = linalg.kron(kin[i], covi_mat[i])
        cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], eff_ind[i + 1][0]:eff_ind[i + 1][-1]] = \
            np.add(cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], eff_ind[i + 1][0]:eff_ind[i + 1][-1]], temp)
    
    # right hand
    rhs_mat = np.divide(rhs_pure, var_com[-1])
    
    # all effects
    cmat = linalg.inv(cmat)
    eff = np.dot(cmat, rhs_mat)
    e = y - xmat.dot(eff[:eff_ind[0][1], :]) - zmat_con.dot(
        eff[eff_ind[0][1]:, :])
    
    return cmat, eff, e


def pre_cov_mat_eigen(cov_dim, var_com):
    start = 0
    end = int(cov_dim * (cov_dim - 1) / 2 + cov_dim)
    cov_add = np.zeros((cov_dim, cov_dim))
    cov_add[np.tril_indices_from(cov_add)] = var_com[start:end]
    cov_add = cov_add + np.tril(cov_add, k=-1).T
    try:
        linalg.cholesky(cov_add)
    except:
        return None
    start = end
    end = 2 * end
    cov_per = np.zeros((cov_dim, cov_dim))
    cov_per[np.tril_indices_from(cov_per)] = var_com[start:end]
    cov_per = cov_per + np.tril(cov_per, k=-1).T
    try:
        linalg.cholesky(cov_per)
    except:
        return None
    if var_com[-1] < 0:
        return None
    else:
        cov_res = np.array([[var_com[-1]]])
    return cov_add, cov_per, cov_res





"""
Analyze the mouse data. Partition the phenotypic variance into addtive + dominance + additve by additive + additive by dominance
+ dominance by dominance + residual

"""
import logging
logging.basicConfig(level=logging.INFO)

# prepare the phenotypic vector, design matrixed for fixed effects and random effects
from gmat.uvlmm import design_matrix_wemai_multi_gmat

pheno_file = '../data/mouse/pheno'
bed_file = '../data/mouse/plink'
y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)

# Calculate the genomic relationship matrix
from gmat.gmatrix.gmatrix import agmat, dgmat_as
a = agmat(bed_file, inv=True)
b = dgmat_as(bed_file, inv=True)

zmat_lst = [zmat, zmat]
gmat_lst = [a[0], b[0]]
gmat_inv_lst = [a[1], b[1]]


from gmat.uvlmm.varcom import em_mme, em_vmat
em_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=10, cc_par=1e-08)
em_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=10, cc_par=1e-08)

from gmat.uvlmm.varcom import em_mme2, em_vmat2
em_mme2(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=10, cc_par=1e-08)
em_vmat2(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=10, cc_par=1e-08)

from gmat.uvlmm.varcom import em_mme22, em_vmat22
em_mme22(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=10, cc_par=1e-08)
em_vmat22(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=10, cc_par=1e-08)


from gmat.uvlmm.varcom import pxem_mme, pxem_vmat
# gmat_lst = [a[0], b[0], a[0] * a[0], a[0]*b[0], b[0]*b[0]]
pxem_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=10, cc_par=1e-08)
pxem_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=10, cc_par=1e-08)

from gmat.uvlmm.varcom import pxem_mme2, pxem_vmat2
pxem_mme2(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1e-08)
pxem_vmat2(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1e-08)

from gmat.uvlmm.varcom import ai_vmat, ai_mme
ai_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1e-08)
ai_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1e-08)

from gmat.uvlmm.varcom import weight_emai_vmat, weight_emai_mme
weight_emai_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1e-08)
weight_emai_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1e-08)


from gmat.uvlmm.varcom import weight_pxemai_vmat, weight_pxemai_mme
zmat_lst = [zmat, zmat, zmat, zmat, zmat]
gmat_lst = [a[0], b[0], a[0]*a[0], a[0]*b[0], b[0]*b[0]]
weight_pxemai_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1e-08)
weight_pxemai_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1e-08)




# Estimate the variances
from gmat.uvlmm.uvlmm_varcom import _wemai_multi_gmat
gmat_lst = [a[0], b[0]]
res = _wemai_multi_gmat(y, xmat, zmat, gmat_lst, init=None, maxiter=200, cc_par=1.0e-8, cc_gra=1.0e-6)



"""
Analyze the yeast data. Partition the phenotypic variance into addtive + dominance + additve by additive + additive by dominance
+ dominance by dominance + individual-specific residual + residual

"""
import numpy as np
import logging
logging.basicConfig(level=logging.INFO)

# prepare the phenotypic vector, design matrixed for fixed effects and random effects
from gmat.uvlmm import design_matrix_wemai_multi_gmat

pheno_file = '../data/yeast/CobaltChloride'
bed_file = '../data/yeast/CobaltChloride'
y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)

# Calculate the genomic relationship matrix
from gmat.gmatrix.gmatrix import agmat
a = agmat(bed_file, inv=False)

# Estimate the variances
from gmat.uvlmm.uvlmm_varcom import wemai_multi_gmat
gmat_lst = [a[0], a[0]*a[0], np.eye(a[0].shape[0])]
res = wemai_multi_gmat(y, xmat, zmat, gmat_lst, init=None, maxiter=200, cc_par=1.0e-8, cc_gra=1.0e-6)






import numpy as np
from scipy import linalg
import pandas as pd
from functools import reduce
import gc
import time
import logging
from tqdm import tqdm
import sys
from scipy.stats import chi2

from gmat.process_plink.process_plink import read_plink, impute_geno
from gmat.uvlmm.design_matrix import design_matrix_wemai_multi_gmat


def _remma_epiAD_pair(y, xmat, zmat, gmat_lst, var_com, bed_file, snp_pair_file, max_test_pair=50000, p_cut=1.0e-4, out_file='epiAD_pair'):
    """
    Given a SNP pair file, perform additive by dominance epistasis test by random SNP-BLUP model.
    :param y: phenotypic vector
    :param xmat: Designed matrix for fixed effect
    :param zmat: csr sparse matrix. Designed matrix for random effect.
    :param gmat_lst: A list for relationship matrix
    :param var_com: Estimated variances
    :param bed_file: the prefix for plink binary file
    :param snp_pair_file: a file containing index for SNP pairs. The program only reads the first two columns and test
    SNP pairs row by row. The max value is num_snp - 1, and the min value is 0.
    :param max_test_pair: The max number of SNP pairs stored in memory. Default value is 50000.
    :param p_cut: put cut value. default value is 0.0001.
    :param out_file: output file. default value is 'remma_epiAA_pair'.
    :return: 0
    """
    logging.info("Calculate the phenotypic covariance matrix and inversion")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += zmat.dot((zmat.dot(gmat_lst[val])).T) * var_com[val]
    del gmat_lst
    gc.collect()
    vmat_inv = np.linalg.inv(vmat)
    logging.info("Calculate P matrix")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = zmat.T.dot(np.dot(pmat, y))
    # pvpmat = reduce(np.dot, [pmat, vmat, pmat])  # pvp = p
    pvpmat = zmat.T.dot((zmat.T.dot(pmat)).T)
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("Read the SNP")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, -1)
    snp_matA = snp_mat - 2 * freq
    snp_mat[snp_mat > 1.5] = 0.0  # 2替换为0, 变为0、1、0编码
    snp_matD = snp_mat - 2 * freq * (1 - freq)
    del snp_mat
    gc.collect()
    logging.info("Test")
    np.savetxt(out_file, ['snp_0 snp_1 eff var chi p'], fmt='%s')
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    ipart = -1
    while True:
        ipart += 1
        skiprows = 1 + ipart * max_test_pair
        try:
            snp_pair = pd.read_csv(snp_pair_file, header=None, sep='\s+', skiprows=skiprows, nrows=max_test_pair)
        except Exception as e:
            logging.info(e)
            break
        snp_pair = np.array(snp_pair.iloc[:, 0:2], dtype=np.int)
        if np.max(snp_pair) > num_snp - 1 or np.min(snp_pair) < 0:
            logging.error('snp_pair is out of range!')
            sys.exit()
        epi_mat = snp_matA[:, snp_pair[:, 0]] * snp_matD[:, snp_pair[:, 1]]
        eff_vec = np.dot(epi_mat.T, pymat)
        var_vec = np.sum(epi_mat * np.dot(pvpmat, epi_mat), axis=0)
        var_vec = var_vec.reshape(-1, 1)
        chi_vec = eff_vec * eff_vec / var_vec
        p_vec = chi2.sf(chi_vec, 1)
        res = pd.DataFrame(
            {0: snp_pair[:, 0], 1: snp_pair[:, 1], 2: eff_vec[:, -1], 3: var_vec[:, -1], 4: chi_vec[:, -1],
             5: p_vec[:, -1]})
        res = res[res[5] < p_cut]
        res.to_csv(out_file, sep=' ', header=False, index=False, mode='a')
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    return 0


def remma_epiAD_pair(pheno_file, bed_file, gmat_lst, var_com, snp_pair_file, max_test_pair=50000, p_cut=1.0e-4, out_file='epiAD_pair'):
    """
    Given a SNP pair file, perform additive by additive epistasis test by random SNP-BLUP model.
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: A list for relationship matrix
    :param var_com: Estimated variances
    :param snp_pair_file: a file containing index for SNP pairs. The program only reads the first two columns and test
    SNP pairs row by row. The max value is num_snp - 1, and the min value is 0.
    :param max_test_pair: The max number of SNP pairs stored in memory. Default value is 50000.
    :param p_cut: put cut value. default value is 0.0001.
    :param out_file: output file. default value is 'epiAD_pair'.
    :return: 0
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)
    res = _remma_epiAD_pair(y, xmat, zmat, gmat_lst, var_com, bed_file, snp_pair_file,
                            max_test_pair=max_test_pair, p_cut=p_cut, out_file=out_file)
    return res



from gmat.longwas.unbalance.unbalance_varcom import unbalance_varcom
from gmat.longwas.unbalance.unbalance_longwas_fixed import unbalance_longwas_fixed
from gmat.longwas.unbalance.unbalance_longwas_fixed_permutation import unbalance_longwas_fixed_permutation
from gmat.longwas.unbalance.unbalance_longwas_trans import unbalance_longwas_trans
from gmat.longwas.unbalance.unbalance_longwas_trans_permutation import unbalance_longwas_trans_permutation



from gmat.longwas.balance.balance_varcom import balance_varcom
from gmat.longwas.balance.balance_longwas_fixed import balance_longwas_fixed
from gmat.longwas.balance.balance_longwas_trans import balance_longwas_trans

from gmat.longwas.balance.balance_longwas_fixed_permutation import balance_longwas_fixed_permutation
from gmat.longwas.balance.balance_longwas_trans_permutation import balance_longwas_trans_permutation



from gmat.longwas.unbalance.unbalance_varcom import unbalance_varcom

from scipy import sparse
from scipy.sparse import csr_matrix, hstack
import numpy as np
import pandas as pd
from patsy import dmatrix
import gc
import os

def leg_mt(time, tmax, tmin, order):
    time = np.array(time, dtype=float).reshape(max(time.shape), 1)
    tvec = 2 * (time - tmin) / (tmax - tmin) - 1
    pmat = []
    for k in range(order + 1):
        c = int(k / 2)
        j = k
        p = 0
        for r in range(0, c + 1):
            p += np.sqrt((2 * j + 1.0) / 2.0) * pow(0.5, j) * (pow(-1, r) *
                                    np.math.factorial(2 * j - 2 * r) / (np.math.factorial(r) *
                                    np.math.factorial(j - r) * np.math.factorial(j - 2 * r))) * pow(tvec, j - 2 * r)
        p = np.array(p)
        pmat.append(p)
    return pmat

from collections import defaultdict
def dct_2D():
    # declare the 2D dict
    return defaultdict(dct_21D)


def dct_21D():
    return defaultdict()



os.chdir("/data/zhanglab/ningc/Acode/GMAT/PKG/examples/data/mouse_long/")

import logging
from gmat.gmatrix import agmat, dgmat_as

logging.basicConfig(level=logging.INFO)

bed_file = 'plink'

# matrix form
agmat0 = agmat(bed_file, inv=True, small_val=0.001, out_fmt='mat')



data_file = 'phe.unbalance.txt'
data_df = pd.read_csv(data_file, sep='\s+', header=0)

na_method = 'omit'
if na_method == 'omit':
    data_df = data_df.dropna()
elif na_method == 'include':
    data_df = data_df.fillna(method='ffill')
    data_df = data_df.fillna(method='bfill')

col_names = data_df.columns

class_vec = []
for val in col_names:
    if not val[0].isalpha():
        print("The first character of columns names must be alphabet!")
        exit()
    if val[0] == val.capitalize()[0]:
        class_vec.append(val)
        data_df[val] = data_df[val].astype('str')
    else:
        try:
            data_df[val] = data_df[val].astype('float')
        except Exception as e:
            print(e)
            print(val, "may contain string, please check!")
            exit()


id_order = []
id = 'ID'
id_arr = list(data_df[id])
id_order.append(id_arr[0])
for i in range(1, len(id_arr)):
    if id_arr[i] != id_arr[i - 1]:
        id_order.append(id_arr[i])

id_in_data = set(data_df[id])

tpoint = 'weak'
trait = 'trait'


code_val = {}
code_dct = dct_2D()
for val in class_vec:
    code_val[val] = 0
    temp = []
    for i in range(data_df.shape[0]):
        if data_df[val][i] not in code_dct[val]:
            code_val[val] += 1
            code_dct[val][data_df[val][i]] = str(code_val[val])
        temp.append(code_dct[val][data_df[val][i]])

    data_df[val] = np.array(temp)

for val in class_vec:
    data_df[val] = data_df[val].astype('int')

forder = 3

leg_fix = leg_mt(data_df[tpoint],1, 16,  forder)
xmat_t = np.concatenate(leg_fix, axis=1)
xmat_t = csr_matrix(xmat_t)


max_id = max(data_df[id]) + 1
tmin = min(data_df[tpoint])
tmax = max(data_df[tpoint])
leg_lst = []  # legendre polynomials for time dependent fixed SNP effects, save for each individuals
for i in range(1, max_id):
    leg_lst.append(leg_mt(data_df[data_df[id] == i][tpoint], tmax, tmin, forder))

aorder = 3
porder = 3

kin_inv = agmat0[1]
leg_add = leg_mt(data_df[tpoint], 1, 16, aorder)
row = np.array(range(data_df.shape[0]))
col = np.array(data_df[id]) - 1
val = np.array([1.0] * data_df.shape[0])
add_mat = csr_matrix((val, (row, col)), shape=(data_df.shape[0], kin_inv.shape[0]))
zmat_add = []
for i in range(len(leg_add)):
    zmat_add.append(add_mat.multiply(leg_add[i]))

leg_per = leg_mt(data_df[tpoint], 1, 16, porder)
per_mat = csr_matrix((val, (row, col)))
zmat_per = []
for i in range(len(leg_per)):
    zmat_per.append((per_mat.multiply(leg_per[i])))

zmat = [zmat_add, zmat_per]
y = data_df[trait].values.reshape(data_df.shape[0], 1)

unbalance_varcom(y, xmat_t, zmat, kin_inv, init=None, max_iter=100, cc_par=1e-08, cc_gra=1e-06, em_weight_step=0.01)



import logging
import sys
import numpy as np
from scipy.sparse import csr_matrix


def design_matrix_wemai_multi_gmat(pheno_file, bed_file):
    """
    prepare the phenotypic vectors, design matrixes of fixed effect and random effect for wemai_multi_gmat program
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :return: phenotypic vectors, design matrixes of fixed effect, design matrixes of random effect (csr_matrix form)
    """
    id_bed_lst = []
    with open(bed_file + '.fam') as fin:
        for line in fin:
            arr = line.split()
            id_bed_lst.append(" ".join([arr[0], arr[1]]))
    id_pheno_lst = {}
    with open(pheno_file) as fin:
        for line in fin:
            arr = line.split()
            if arr[-1] not in ['NA', 'NaN', 'nan', 'na']:
                try:
                    id_pheno_lst[" ".join([arr[0], arr[1]])].append(" ".join(arr))
                except Exception as e:
                    del e
                    id_pheno_lst[" ".join([arr[0], arr[1]])] = [" ".join(arr)]
    id_not_pheno = set(id_bed_lst) - set(list(id_pheno_lst.keys()))
    if len(id_not_pheno) > 0:
        logging.error('The below genotyped id is not in the phenotype file:\n {}'.format('\n'.join(list(id_not_pheno))))
        sys.exit()
    y = []
    xmat = []
    id_lst = []
    for id in id_bed_lst:
        for val in id_pheno_lst[id]:
            arr = val.split()
            y.append(float(arr[-1]))
            xmat.append(arr[2:-1])
            id_lst.append(arr[1])
    y = np.array(y).reshape(-1, 1)
    xmat = np.array(xmat, dtype=float).reshape(y.shape[0], -1)
    id_dct = {}
    row = []
    col = []
    j = 0
    for i in range(len(id_lst)):
        row.append(i)
        if id_lst[i] not in id_dct:
            id_dct[id_lst[i]] = j
            j += 1
        col.append(id_dct[id_lst[i]])
    zmat = csr_matrix(([1.0]*len(row), (row, col)))
    return y, xmat, zmat


def design_matrix_wemai_multi_gmat_pred(pheno_file, bed_file):
    """
    prepare the phenotypic vectors, design matrixes of fixed effect and random effect for wemai_multi_gmat program
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :return: phenotypic vectors, design matrixes of fixed effect, design matrixes of random effect (csr_matrix form)
    """
    id_bed_lst = []
    with open(bed_file + '.fam') as fin:
        for line in fin:
            arr = line.split()
            id_bed_lst.append(" ".join([arr[0], arr[1]]))
    id_pheno_lst = {}
    with open(pheno_file) as fin:
        for line in fin:
            arr = line.split()
            if arr[-1] not in ['NA', 'NaN', 'nan', 'na']:
                try:
                    id_pheno_lst[" ".join([arr[0], arr[1]])].append(" ".join(arr))
                except Exception as e:
                    del e
                    id_pheno_lst[" ".join([arr[0], arr[1]])] = [" ".join(arr)]
    y = []
    xmat = []
    id_lst = []
    for id in id_bed_lst:
        if id in id_pheno_lst:
            for val in id_pheno_lst[id]:
                arr = val.split()
                y.append(float(arr[-1]))
                xmat.append(arr[2:-1])
                id_lst.append(arr[1])
        else:
            id_lst.append('NA')
    y = np.array(y).reshape(-1, 1)
    xmat = np.array(xmat, dtype=float).reshape(y.shape[0], -1)
    id_dct = {}
    row, col, val = [], [], []
    rowi, coli = 0, 0
    for i in range(len(id_lst)):
        if id_lst[i] != 'NA':
            row.append(rowi)
            rowi += 1
            if id_lst[i] not in id_dct:
                id_dct[id_lst[i]] = coli
                coli += 1
            col.append(id_dct[id_lst[i]])
            val.append(1.0)
        else:
            coli += 1
    zmat = csr_matrix((val, (row, col)), shape=(rowi, coli))
    return y, xmat, zmat



import sys
import logging
import numpy as np
from scipy import linalg
import pandas as pd
import time
from gmat.process_plink.process_plink import read_plink, impute_geno


def output_mat(mat, id, out_file, out_fmt):
    if out_fmt == 'mat':
        np.savetxt(out_file + '0', mat)
    elif out_fmt == 'row_col_val':
        ind = np.tril_indices_from(mat)
        df = pd.DataFrame({
            "row": ind[0]+1,
            "col": ind[1]+1,
            "val": mat[ind]
        })
        df.to_csv(out_file + '1', sep=' ', index=False, header=False, columns=["row", "col", "val"])
    elif out_fmt == 'id_id_val':
        ind = np.tril_indices_from(mat)
        df = pd.DataFrame({
            "id0": id[ind[0]],
            "id1": id[ind[1]],
            "val": mat[ind]
        })
        df.to_csv(out_file + '2', sep=' ', index=False, header=False, columns=["id0", "id1", "val"])
    else:
        return 0
    return 1


def agmat(bed_file, inv=False, small_val=0.001, out_fmt='mat'):
    """
    additive genomic relationship matrix and its inversion
    :param bed_file: The prefix for plink binary file
    :param inv: Whether to calculate the inversion. Default value is True
    :param small_val: A small vale added to the diagonal to grant the positive definite. Default value is 0.001.
    :param out_fmt: the output format. mat: matrix format (default); row_col_val: row-column-value format;
    id_id_val: id-id-value format.
    :return: return numpy array for genomic relationship matrix and its inversion. Output the matrixes into the file
    with prefix of bed_file.
    """
    logging.info("{:#^80}".format("Read the SNP data"))
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        logging.info('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    num_id = snp_mat.shape[0]  # 个体数
    num_snp = snp_mat.shape[1]  # SNP数
    logging.info("There are {:d} individuals and {:d} SNPs.".format(num_id, num_snp))
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    scale = 2 * freq * (1 - freq)  # 标准化因子
    scale = np.sum(scale)
    logging.info('The scaled factor is: {:.3f}'.format(scale))
    snp_mat = snp_mat - 2 * freq
    
    logging.info("{:#^80}".format("Calculate the additive genomic relationship matrix"))
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    kin = np.dot(snp_mat, snp_mat.T) / scale
    kin_diag = np.diag(kin)
    kin_diag = kin_diag + kin_diag * small_val
    np.fill_diagonal(kin, kin_diag)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    
    logging.info("{:#^80}".format("Output"))
    fam_info = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id = np.array(fam_info.iloc[:, 1])
    out_file = bed_file + '.agrm'
    logging.info("The output file is " + out_file)
    res = output_mat(kin, id, out_file, out_fmt)
    if res == 0:
        logging.error('Not Recognized output format: ' + out_fmt)
        sys.exit()
    kin_inv = None
    if inv:
        logging.info("{:#^80}".format("Calculate the inversion of kinship"))
        clock_t0 = time.perf_counter()
        cpu_t0 = time.process_time()
        kin_inv = linalg.inv(kin)
        clock_t1 = time.perf_counter()
        cpu_t1 = time.process_time()
        logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0,
                                                                                   cpu_t1 - cpu_t0))
        logging.info("{:#^80}".format("Output the inversion"))
        out_file = bed_file + '.agiv'
        logging.info("The output file is: " + out_file)
        output_mat(kin_inv, id, out_file, out_fmt)
    return kin, kin_inv


def dgmat_as(bed_file, inv=False, small_val=0.001, out_fmt='mat'):
    """
    dominance genomic relationship matrix and its inversion
    :param bed_file: The prefix for plink binary file
    :param inv: Whether to calculate the inversion. Default value is True
    :param small_val: A small vale added to the diagonal to grant the positive definite. Default value is 0.001.
    :param out_fmt: the output format. mat: matrix format (default); row_col_val: row-column-value format;
    id_id_val: id-id-value format.
    :return: return numpy array for genomic relationship matrix and its inversion. Output the matrixes into the file
    with prefix of bed_file.
    """
    logging.info("{:#^80}".format("Read, impute and scale the SNP"))
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        print('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    num_id = snp_mat.shape[0]  # 个体数
    num_snp = snp_mat.shape[1]  # SNP数
    logging.info("There are {:d} individuals and {:d} SNPs.".format(num_id, num_snp))
    freq = np.sum(snp_mat, axis=0) / (2 * snp_mat.shape[0])
    freq.shape = (1, num_snp)
    scale_vec = 2 * freq * (1 - freq)
    scale = np.sum(scale_vec * (1 - scale_vec))
    logging.info('The scaled factor is: {:.3f}'.format(scale))
    snp_mat[snp_mat > 1.5] = 0.0  # 2替换为0, 变为0、1、0编码
    snp_mat = snp_mat - scale_vec

    logging.info("{:#^80}".format("Calculate the dominance genomic relationship matrix"))
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    kin = np.dot(snp_mat, snp_mat.T) / scale
    kin_diag = np.diag(kin)
    kin_diag = kin_diag + kin_diag * small_val
    np.fill_diagonal(kin, kin_diag)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info(
        "Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))

    logging.info("{:#^80}".format("Output"))
    fam_info = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id = np.array(fam_info.iloc[:, 1])
    out_file = bed_file + '.dgrm_as'
    logging.info("The output file is " + out_file)
    res = output_mat(kin, id, out_file, out_fmt)
    if res == 0:
        logging.error('Not Recognized output format: ' + out_fmt)
        sys.exit()
    kin_inv = None
    if inv:
        logging.info("{:#^80}".format("Calculate the inversion of kinship"))
        clock_t0 = time.perf_counter()
        cpu_t0 = time.process_time()
        kin_inv = linalg.inv(kin)
        clock_t1 = time.perf_counter()
        cpu_t1 = time.process_time()
        logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0,
                                                                                   cpu_t1 - cpu_t0))
        logging.info("{:#^80}".format("Output the inversion"))
        out_file = bed_file + '.dgiv_as'
        logging.info("The output file is: " + out_file)
        output_mat(kin_inv, id, out_file, out_fmt)
    return kin, kin_inv


def ginbreedcoef(bed_file):
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        print('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    print("There are {:d} individuals and {:d} SNPs.".format(snp_mat.shape[0], snp_mat.shape[1]))
    homo_f = 1 - np.sum(np.abs(snp_mat - 1.0) < 0.01, axis=1) / snp_mat.shape[1]
    freq = np.sum(snp_mat, axis=0) / (2 * snp_mat.shape[0])
    freq.shape = (1, snp_mat.shape[1])
    scale_vec = 2 * freq * (1 - freq)
    scale = np.sum(scale_vec)
    snp_mat = snp_mat - 2 * freq
    grm_f1 = np.sum(np.multiply(snp_mat, snp_mat), axis=1) / scale - 1.0
    grm_f2 = np.sum(np.multiply(snp_mat, snp_mat) / scale_vec, axis=1) / snp_mat.shape[1] - 1.0
    
    fam_info = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id = np.array(fam_info.iloc[:, 1])
    data_df = {'id': id, 'homo_F': homo_f, 'grm_F1': grm_f1, 'grm_F2': grm_f2}
    data_df = pd.DataFrame(data_df, columns=['id', 'homo_F', 'grm_F1', 'grm_F2'])
    out_file = bed_file + '.ginbreedcoef'
    data_df.to_csv(out_file, sep=' ', header=True, index=False)




from setuptools import setup, find_packages

setup(
    cffi_modules=["gmat/remma/_build.py:ffi", "gmat/process_plink/_build.py:ffi"],
    name='gmat',
    version='2020.12.16',
    description='Genomic Multivariate Analysis Tools',
    # long_description=open('README.rst').read(),
    # long_description_content_type="text/markdown",
    author='Chao Ning',
    author_email='ningchao91@gmail.com',
    packages=find_packages(),
    platforms=["all"],
    url="https://github.com/chaoning/GMAT",
    include_package_data = True,
    install_requires=[
        'numpy>=1.16.0',
        'pandas>=0.19.0',
        'scipy>=1.1.1',
        'cffi>=1.12.0',
        'pandas_plink>=2.0.0',
        #'pysnptools>=0.4.11',
        'tqdm>=4.43.0',
    ],
)



import numpy as np
import pandas as pd
from scipy import linalg
import logging
from tqdm import tqdm
from functools import reduce
from gmat.uvlmm.design_matrix import design_matrix_wemai_multi_gmat
from gmat.process_plink.process_plink import read_plink, impute_geno


def lm_snp_eff(pheno_file, bed_file, out_file='lm_snp_eff'):
    """
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param out_file: the output file
    :return: the estimated snp effect
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    snp_eff = []
    for i in tqdm(range(snp_mat.shape[1])):
        xmati = np.concatenate([xmat, snp_mat[:, i:(i+1)]], axis=1)
        eff = np.dot(linalg.inv(np.dot(xmati.T, xmati)), np.dot(xmati.T, y))
        snp_eff.append(eff[-1, -1])
    df = pd.read_csv(bed_file + '.bim', sep='\s+', header=None)
    df['eff'] = snp_eff
    df.to_csv(out_file, sep=' ', header=False, index=False)


def lm_pred(pheno_file, bed_file, agmat, out_file='lm_pred'):
    """
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param agmat: additive genomic relationship matrix
    :param out_file: the output file
    :return: the estimated snp effect
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)
    vmat = np.diag([1] * y.shape[0])
    vxmat = np.dot(vmat, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat - pmat
    zpymat = zmat.T.dot(np.dot(pmat, y))
    eff = np.dot(agmat, zpymat)
    np.savetxt(out_file + '.rand_eff', eff)



import numpy as np
from scipy import linalg
import pandas as pd
from functools import reduce
import gc
import time
import logging
from tqdm import tqdm
import sys
import os
from scipy.stats import chi2

from gmat.process_plink.process_plink import read_plink, impute_geno
from gmat.uvlmm.design_matrix import design_matrix_wemai_multi_gmat


from _cremma_epi_eff_cpu import ffi, lib


def _remma_epiAA_maf_eff(y, xmat, zmat, gmat_lst, var_com, bed_file, snp_lst_0=None, freq=None, freq_deno=None, p_cut=1.0e-5, out_file='epiAA_eff'):
    """
    Estimate additive by additive epistasis effects by random SNP-BLUP model.
    :param y: phenotypic vector
    :param xmat: Designed matrix for fixed effect
    :param zmat: csr sparse matrix. Designed matrix for random effect.
    :param gmat_lst: A list for relationship matrix
    :param var_com: Estimated variances
    :param bed_file: the prefix for plink binary file
    :param snp_lst_0: the first SNP list for the SNP pairs. the min value is 0 and the max value is num_snp-2. The
    default value is None, which means list [0, num_snp-1)
    :param var_app: the approximate variances for estimated SNP effects.
    :param p_cut: put cut value. default value is 1.0e-5.
    :param out_file: output file. default value is 'epiAA_eff'.
    :return: 0
    """
    logging.info("Calculate the phenotypic covariance matrix and inversion")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += zmat.dot((zmat.dot(gmat_lst[val])).T) * var_com[val]
    del gmat_lst
    gc.collect()
    vmat_inv = np.linalg.inv(vmat)
    logging.info("Calculate P matrix")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = zmat.T.dot(np.dot(pmat, y))
    del vmat, vmat_inv, pmat
    gc.collect()
    num_snp = pd.read_csv(bed_file+'.bim', header=None).shape[0]
    num_id = pd.read_csv(bed_file+'.fam', header=None).shape[0]
    if snp_lst_0 is None:
        snp_lst_0 = range(num_snp - 1)
    else:
        if max(snp_lst_0) >= num_snp - 1 or min(snp_lst_0) < 0:
            logging.error('snp_lst_0 is out of range!')
            sys.exit()
    if freq is None:
        freq = np.zeros((num_snp,), dtype=np.longlong)
    if freq_deno is None:
        freq_deno = np.ones(111)
    logging.info("Convert python variates to C type")
    pbed_file = ffi.new("char[]", bed_file.encode('ascii'))
    pnum_id = ffi.cast("long long", num_id)
    pnum_snp = ffi.cast("long long", num_snp)
    snp_lst_0 = np.array(list(snp_lst_0), dtype=np.longlong)
    psnp_lst_0 = ffi.cast("long long *", snp_lst_0.ctypes.data)
    # psnp_lst_0 = ffi.cast("long long *", ffi.from_buffer(snp_lst_0))
    plen_snp_lst_0 = ffi.cast("long long", len(snp_lst_0))
    pfreq = ffi.cast("long long *", freq.ctypes.data)
    ppymat = ffi.cast("double *", pymat.ctypes.data)
    chi_cut = chi2.isf(p_cut, 1)
    eff_cut = np.array(np.sqrt(chi_cut * freq_deno))
    np.savetxt('eff_cut', eff_cut)
    peff_cut = ffi.cast("double *", eff_cut.ctypes.data)
    # peff_cut = ffi.cast("double", 1.0)
    temp_file = out_file + '.temp'
    pout_file = ffi.new("char[]", temp_file.encode('ascii'))
    logging.info('Test')
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    lib.remma_epiAA_maf_eff_cpu(pbed_file, pnum_id, pnum_snp, psnp_lst_0, plen_snp_lst_0, ppymat, pfreq, peff_cut, pout_file)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    logging.info('Add the approximate P values')
    # var_app = 1.0  ################
    with open(temp_file) as fin, open(out_file, 'w') as fout:
        head_line = fin.readline()
        head_line = head_line.strip()
        head_line += ' chi_app p_app\n'
        fout.write(head_line)
        for line in fin:
            arr = line.split()
            chi_app = float(arr[-1]) * float(arr[-1]) / freq_deno[freq[int(arr[0])] * 10 + freq[int(arr[1])]]
            p_app = chi2.sf(chi_app, 1)
            fout.write(' '.join(arr + [str(chi_app), str(p_app)]) + '\n')
    os.remove(temp_file)
    return 0


def remma_epiAA_maf_eff(pheno_file, bed_file, gmat_lst, var_com, snp_lst_0=None, freq=None, freq_deno=None, p_cut=1.0e-5, out_file='epiAA_maf_eff'):
    """
    Estimate additive by additive epistasis effects by random SNP-BLUP model.
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: A list for relationship matrix
    :param var_com: Estimated variances
    :param snp_lst_0: the first SNP list for the SNP pairs. the min value is 0 and the max value is num_snp-2. The
    default value is None, which means list [0, num_snp-1)
    :param var_app: the approximate variances for estimated SNP effects.
    :param p_cut: put cut value. default value is 1.0e-5.
    :param out_file: output file. default value is 'epiAA_eff'.
    :return: 0
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)
    res = _remma_epiAA_maf_eff(y, xmat, zmat, gmat_lst, var_com, bed_file,
                           snp_lst_0=snp_lst_0, freq=freq, freq_deno=freq_deno, p_cut=p_cut, out_file=out_file)
    return res


def _remma_epiAA_maf_eff_parallel(y, xmat, zmat, gmat_lst, var_com, bed_file, parallel,
                                  freq=None, freq_deno=None, p_cut=1.0e-5, out_file='epiAA_maf_eff_parallel'):
    """
    Parallel version. Additive by additive epistasis test by random SNP-BLUP model.
    :param y: phenotypic vector
    :param xmat: Designed matrix for fixed effect
    :param zmat: csr sparse matrix. Designed matrix for random effect.
    :param gmat_lst: A list for relationship matrix
    :param var_com: Estimated variances
    :param bed_file: the prefix for plink binary file
    :param parallel: A list containing two integers. The first integer is the number of parts to parallel. The second
    integer is the part to run. For example, parallel = [3, 1], parallel = [3, 2] and parallel = [3, 3] mean to divide
    total number of tests into three parts and run parallelly.
    :param var_app: the approximate variances for estimated SNP effects.
    :param p_cut: put cut value. default value is 1.0e-5.
    :param out_file: output file. default value is 'remma_epiAA_eff_parallel'.
    :return: 0
    """
    logging.info("Parallel: " + str(parallel[0]) + ', ' + str(parallel[1]))
    bim_df = pd.read_csv(bed_file + '.bim', header=None)
    num_snp = bim_df.shape[0]
    num_snp_part = int(num_snp/(2*parallel[0]))
    snp_pos_0 = (parallel[1]-1) * num_snp_part
    snp_pos_1 = parallel[1] * num_snp_part
    snp_pos_2 = (2*parallel[0] - parallel[1]) * num_snp_part
    snp_pos_3 = (2*parallel[0] - parallel[1] + 1) * num_snp_part
    if parallel[1] == 1:
        snp_pos_3 = num_snp - 1
    logging.info('SNP position point: ' +
                 ','.join(list(np.array([snp_pos_0, snp_pos_1, snp_pos_2, snp_pos_3], dtype=str))))
    snp_list_0 = list(range(snp_pos_0, snp_pos_1)) + list(range(snp_pos_2, snp_pos_3))
    res = _remma_epiAA_maf_eff(y, xmat, zmat, gmat_lst, var_com, bed_file, snp_lst_0=snp_list_0, freq=freq,
                               freq_deno=freq_deno, p_cut=p_cut, out_file=out_file + '.' + str(parallel[1]))
    return res


def remma_epiAA_maf_eff_parallel(pheno_file, bed_file, gmat_lst, var_com, parallel,
                                   freq=None, freq_deno=None,  p_cut=1.0e-5, out_file='epiAA_maf_eff_parallel'):
    """
    Parallel version. Additive by additive epistasis test by random SNP-BLUP model.
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: A list for relationship matrix
    :param var_com: Estimated variances
    :param parallel: A list containing two integers. The first integer is the number of parts to parallel. The second
    integer is the part to run. For example, parallel = [3, 1], parallel = [3, 2] and parallel = [3, 3] mean to divide
    :param var_app: the approximate variances for estimated SNP effects.
    :param p_cut: put cut value. default value is 1.0e-5.
    :param out_file: output file. default value is 'epiAA_maf_eff_parallel'.
    :return: 0
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)
    res = _remma_epiAA_maf_eff_parallel(y, xmat, zmat, gmat_lst, var_com, bed_file, parallel,
                                   freq=freq, freq_deno=freq_deno,  p_cut=p_cut, out_file=out_file)
    return res




import math
import numpy as np
import pandas as pd
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, isspmatrix, hstack, vstack
import datetime
import logging

from .common import *
from .pre_mat import *
from .iter_mat import *


def unbalance_emai(y, xmat, zmat, kin, init=None, maxiter=30, cc_par=1.0e-8, cc_gra=1.0e-6, em_weight_step=0.001):
    num_var_pos = 1
    for i in range(len(zmat)):
        num_var_pos += len(zmat[i])
    y_var = np.var(y)/num_var_pos
    num_record = y.shape[0]
    var_com = []
    eff_ind = [[0, xmat.shape[1]]]  # the index for all effects [start end]
    zmat_con_lst = []  # combined random matrix
    cov_dim = []  # the dim for covariance matrix
    vari = []
    varij = []
    varik = []
    i = 0
    for i in range(len(zmat)):
        temp = [eff_ind[i][-1]]
        zmat_con_lst.append(hstack(zmat[i]))
        cov_dim.append(len(zmat[i]))
        for j in range(len(zmat[i])):
            temp.append(temp[-1] + zmat[i][j].shape[1])
            for k in range(j + 1):
                vari.append(i+1)
                varij.append(j+1)
                varik.append(k+1)
                if j == k:
                    var_com.append(y_var)
                else:
                    var_com.append(0.0)
        eff_ind.append(temp)
    var_com.append(y_var)
    vari.append(i + 2)
    varij.append(1)
    varik.append(1)
    if init is None:
        var_com = np.array(var_com)
    else:
        if len(var_com) != len(init):
            logging.info('ERROR: The length of initial variances should be' + len(var_com))
            exit()
        else:
            var_com = np.array(init)
    var_com_update = np.array(var_com)
    logging.info('***prepare the MME**')
    zmat_con = hstack(zmat_con_lst)  # design matrix for random effects
    wmat = hstack([xmat, zmat_con])  # merged design matrix
    cmat_pure = np.dot(wmat.T, wmat)  # C matrix
    rhs_pure = wmat.T.dot(y)  # right hand
    # em weight vector
    if em_weight_step <= 0.0 or em_weight_step > 1.0:
        logging.info('ERROR: The em weight step should be between 0 (not include) and 1 (include)')
        exit()
    iter_count = 0
    cc_par_val = 1000.0
    cc_gra_val = 1000.0
    delta = 1000.0
    logging.info("initial variances: " + ' '.join(np.array(var_com, dtype=str)))
    covi_mat = pre_covi_mat(cov_dim, var_com)
    if covi_mat is None:
        logging.inf("ERROR: Initial variances is not positive define, please check!")
        exit()
    while iter_count < maxiter:
        iter_count += 1
        logging.info('***Start the iteration: ' + str(iter_count) + ' ***')
        logging.info("Prepare the coefficient matrix")
        cmat = (cmat_pure.multiply(1.0 / var_com[-1])).toarray()
        for i in range(len(cov_dim)):
            if isspmatrix(kin[i]):
                temp = sparse.kron(covi_mat[i], kin[i])
                temp = temp.toarray()
            else:
                temp = linalg.kron(covi_mat[i], kin[i])
            cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
            eff_ind[i + 1][0]:eff_ind[i + 1][-1]] = \
                np.add(cmat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
                       eff_ind[i + 1][0]:eff_ind[i + 1][-1]], temp)
        rhs_mat = np.divide(rhs_pure, var_com[-1])
        cmati = linalg.inv(cmat)
        eff = np.dot(cmati, rhs_mat)
        e = y - xmat.dot(eff[:eff_ind[0][1], :]) - zmat_con.dot(
            eff[eff_ind[0][1]:, :])
        # first-order derivative
        fd_mat = pre_fd_mat_x(cmati, kin, covi_mat, eff, eff_ind, e, cov_dim, zmat_con_lst, wmat, num_record, var_com)
        # AI matrix
        ai_mat = pre_ai_mat(cmati, covi_mat, eff, eff_ind, e, cov_dim, zmat_con_lst, wmat, var_com)
        # EM matrix
        em_mat = pre_em_mat(cov_dim, zmat_con_lst, num_record, var_com)
        # Increase em weight to guarantee variances positive
        gamma = -em_weight_step
        while gamma < 1.0:
            gamma = gamma + em_weight_step
            if gamma >= 1.0:
                gamma = 1.0
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_update = var_com + delta
            covi_mat = pre_covi_mat(cov_dim, var_com_update)
            if covi_mat is not None:
                logging.info('EM weight value: ' + str(gamma))
                break
        logging.info('Updated variances: ' + ' '.join(np.array(var_com_update, dtype=str)))
        if covi_mat is None:
            logging.info("ERROR: Updated variances is not positive define!")
            exit()
        # Convergence criteria
        cc_par_val = np.sum(pow(delta, 2)) / np.sum(pow(var_com_update, 2))
        cc_par_val = np.sqrt(cc_par_val)
        cc_gra_val = np.sqrt(np.sum(pow(fd_mat, 2))) / len(var_com)
        var_com = var_com_update.copy()
        logging.info("Change in parameters, Norm of gradient vector: " + str(cc_par_val) + ', ' + str(cc_gra_val))
        if cc_par_val < cc_par and cc_gra_val < cc_gra:
            break
    if cc_par_val < cc_par and cc_gra_val < cc_gra:
        logging.info("Variances Converged")
    else:
        logging.info("Variances not Converged")
    var_pd = {'vari': vari,
              "varij": varij,
              "varik": varik,
              "var_val": var_com}
    var_pd = pd.DataFrame(var_pd, columns=['vari', "varij", "varik", "var_val"])
    return var_pd



import numpy as np
from scipy import linalg


def uvlmm_varcom_eigen(y, xmat, gmat, init=None, maxiter=100, cc=1.0e-8):
    ''''
    描述：单性状方差组分估计；特征分解加快速度
    '''
    # 特征分解基因组关系矩阵
    gmat_eigen_val, gmat_eigen_vec = linalg.eigh(gmat)
    gmat_eigen_val = gmat_eigen_val.reshape(len(gmat_eigen_val), 1)
    y = np.dot(gmat_eigen_vec.T, y) # 表型旋转
    xmat = np.dot(gmat_eigen_vec.T, xmat) # 固定效应设计矩阵旋转
    if init is not None: # 方差组分迭代初始值
        var = np.array(init)
    else:
        var = np.array([np.var(y)/2]*2)
    fd_mat = np.zeros(2)
    ai_mat = np.zeros((2, 2))
    em_mat = np.zeros((2, 2))
    num_id = gmat.shape[0]
    # 开始估计方差组分
    for i in range(maxiter):
        print('迭代次数：', i + 1)
        vmat = 1.0 / (gmat_eigen_val * var[0] + var[1])
        vx = np.multiply(vmat, xmat)
        xvx = np.dot(xmat.T, vx)
        xvx = np.linalg.inv(xvx)
        # py
        xvy = np.dot(vx.T, y)
        y_xb = y - np.dot(xmat, np.dot(xvx, xvy))
        py = np.multiply(vmat, y_xb)
        # add_py p_add_py
        add_py = np.multiply(gmat_eigen_val, py)
        xvy = np.dot(vx.T, add_py)
        y_xb = add_py - np.dot(xmat, np.dot(xvx, xvy))
        p_add_py = np.multiply(vmat, y_xb)
        # res_py p_res_py
        res_py = py.copy()
        xvy = np.dot(vx.T, res_py)
        y_xb = res_py - np.dot(xmat, np.dot(xvx, xvy))
        p_res_py = np.multiply(vmat, y_xb)
        # fd
        tr_vd = np.sum(np.multiply(vmat, gmat_eigen_val))
        xvdvx = np.dot(xmat.T, vmat * gmat_eigen_val * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, add_py))
        fd_mat[0] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        tr_vd = np.sum(vmat)
        xvdvx = np.dot(xmat.T, vmat * vx)
        tr_2d = np.sum(np.multiply(xvdvx, xvx))
        ypvpy = np.sum(np.dot(py.T, res_py))
        fd_mat[1] = 0.5 * (-tr_vd + tr_2d + ypvpy)
        # AI
        ai_mat[0, 0] = np.sum(np.dot(add_py.T, p_add_py))
        ai_mat[0, 1] = ai_mat[1, 0] = np.sum(np.dot(add_py.T, p_res_py))
        ai_mat[1, 1] = np.sum(np.dot(res_py.T, p_res_py))
        ai_mat = 0.5 * ai_mat
        # EM
        em_mat[0, 0] = num_id / (var[0] * var[0])
        em_mat[1, 1] = num_id / (var[1] * var[1])
        for j in range(0, 51):
            gamma = j * 0.02
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_update = var + delta
            if min(var_update) > 0:
                print('EM weight value:', gamma)
                break
        print('Updated variances:', var_update)
        # Convergence criteria
        cc_val = np.sum(pow(delta, 2)) / np.sum(pow(var_update, 2))
        cc_val = np.sqrt(cc_val)
        var = var_update.copy()
        print("CC: ", cc_val)
        if cc_val < cc:
            break
    return [var, gmat_eigen_vec, gmat_eigen_val]



"""
additive by additive test by SNP-BLUP model. Mouse data.

"""
import logging
import numpy as np
from gmat.uvlmm import design_matrix_wemai_multi_gmat
from gmat.uvlmm import wemai_multi_gmat
from gmat.gmatrix import agmat, dgmat_as
from gmat.remma.remma_cpu import remma_add_cpu

logging.basicConfig(level=logging.INFO)

# prepare the phenotypic vector, design matrixed for fixed effects and random effects
pheno_file = '../data/mouse/pheno'
bed_file = '../data/mouse/plink'
y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)

# Calculate the genomic relationship matrix
a = agmat(bed_file, inv=False)
b = dgmat_as(bed_file, inv=False)


# Example 1: Just include additive relationship matrix
gmat_lst = [a[0]]
var_com_a = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a, bed_file, out_file='../data/mouse/remma_add_cpu_a')


# Example 2: Include additive relationship matrix and dominance relationship matrix
gmat_lst = [a[0], b[0]]
var_com_a_d = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a_d = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a_d, bed_file, out_file='../data/mouse/remma_add_cpu_a_d')


# Example 3: Include additive relationship matrix and additive by additive relationship matrix
gmat_lst = [a[0], a[0]*a[0]]
var_com_a_axa = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a_axa = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a_axa, bed_file, out_file='../data/mouse/remma_add_cpu_a_axa')


# Example 4: Include additive, dominance, additive by additive, additive by dominance,
# dominance by dominance relationship matrix
gmat_lst = [a[0], b[0], a[0]*a[0], a[0]*b[0], b[0]*b[0]]
var_com_all = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_all = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_all, bed_file, out_file='../data/mouse/remma_add_cpu_all')

print('Median values: ', np.median(res_a['p_val']), np.median(res_a_d['p_val']), np.median(res_a_axa['p_val']),
      np.median(res_all['p_val']))
print('Pearson product-moment correlation coefficients: ', np.corrcoef([res_a['p_val'], res_a_d['p_val'], res_a_axa['p_val'], res_all['p_val']]))


"""
additive by additive test by SNP-BLUP model. Yeast data. No heterozygous genotype. No dominance. 
Include repeated measures.

"""

import logging
import numpy as np
from gmat.uvlmm import design_matrix_wemai_multi_gmat
from gmat.uvlmm import wemai_multi_gmat
from gmat.gmatrix import agmat, dgmat_as
from gmat.remma.remma_cpu import remma_add_cpu

logging.basicConfig(level=logging.INFO)

# prepare the phenotypic vector, design matrixed for fixed effects and random effects
pheno_file = '../data/yeast/CobaltChloride'
bed_file = '../data/yeast/CobaltChloride'
y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)

# Calculate the genomic relationship matrix
a = agmat(bed_file, inv=False)

# Example 0: Include additive relationship matrix
gmat_lst = [a[0]]  # np.eye(a[0].shape[0]) is a identity matrix to model the permanent environmental effect
var_com_a = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a, bed_file, out_file='../data/yeast/remma_add_cpu_a')


# Example 1: Include additive relationship matrix and individual-specific errors (Permanent environmental effect)
gmat_lst = [a[0], np.eye(a[0].shape[0])]  # np.eye(a[0].shape[0]) is a identity matrix to model the permanent environmental effect
var_com_a_p = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a_p = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a_p, bed_file, out_file='../data/yeast/remma_add_cpu_a_p')


# Example 2: Include additive, additive by additive relationship matrix
# and individual-specific errors (Permanent environmental effect)

gmat_lst = [a[0], a[0]*a[0], np.eye(a[0].shape[0])]  # np.eye(a[0].shape[0]) is a identity matrix to model the permanent environmental effect
var_com_a_axa_p = wemai_multi_gmat(y, xmat, zmat, gmat_lst)
res_a_axa_p = remma_add_cpu(y, xmat, zmat, gmat_lst, var_com_a_axa_p, bed_file,
                            out_file='../data/yeast/remma_add_cpu_a_axa_p')
